import unittest
import warnings
import pandas as pd
import pyterrier as pt


class TestValidate(unittest.TestCase):
    def test_columns_valid(self):
        with self.subTest('df'):
            pt.validate.columns(
                pd.DataFrame({'qid': [1, 2], 'text': ['hello', 'world']}),
                includes=['qid'],
                excludes=['docno'])
        with self.subTest('cols'):
            pt.validate.columns(
                ['qid', 'text'],
                includes=['qid'],
                excludes=['docno'])
        with self.subTest('iterdict'):
            pt.validate.columns_iter(
                pt.utils.peekable([{'qid': '1', 'text': '0'}]),
                includes=['qid'],
                excludes=['docno'])

    def test_columns_missing_column(self):
        with self.subTest('df'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.columns(
                    pd.DataFrame({'text': ['hello', 'world']}),
                    includes=['qid'])
            self.assertIn('qid', cm.exception.modes[0].missing_columns)
        with self.subTest('cols'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.columns(
                    ['text', 'other'],
                    includes=['qid'])
            self.assertIn('qid', cm.exception.modes[0].missing_columns)
        with self.subTest('iterdict'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.columns_iter(
                    pt.utils.peekable([{'text': 'hello', 'other': 'world'}]),
                    includes=['qid'])
            self.assertIn('qid', cm.exception.modes[0].missing_columns)

    def test_columns_extra_column(self):
        with self.subTest('df'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.columns(
                    pd.DataFrame({'qid': [1, 2], 'docno': [3, 4]}),
                    includes=['qid'],
                    excludes=['docno'])
            self.assertIn('docno', cm.exception.modes[0].extra_columns)
        with self.subTest('cols'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.columns(
                    ['qid', 'docno'],
                    includes=['qid'],
                    excludes=['docno'])
            self.assertIn('docno', cm.exception.modes[0].extra_columns)
        with self.subTest('iterdict'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.columns_iter(
                    pt.utils.peekable([{'qid': '1', 'docno': '2'}]),
                    includes=['qid'],
                    excludes=['docno'])
            self.assertIn('docno', cm.exception.modes[0].extra_columns)

    def test_columns_warn_mode(self):
        with self.subTest('df'):
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                pt.validate.columns(pd.DataFrame({'text': ['hello', 'world']}), includes=['qid'], warn=True)
                self.assertEqual(len(w), 1)
                self.assertTrue(issubclass(w[0].category, pt.validate.InputValidationWarning))
                self.assertIn("does not match required columns", str(w[0].message))
        with self.subTest('cols'):
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                pt.validate.columns(['text', 'other'], includes=['qid'], warn=True)
                self.assertEqual(len(w), 1)
                self.assertTrue(issubclass(w[0].category, pt.validate.InputValidationWarning))
                self.assertIn("does not match required columns", str(w[0].message))
        with self.subTest('iterdict'):
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                pt.validate.columns_iter(
                    pt.utils.peekable([{'text': 'hello', 'other': 'world'}]),
                    includes=['qid'],
                    warn=True)
                self.assertEqual(len(w), 1)
                self.assertTrue(issubclass(w[0].category, pt.validate.InputValidationWarning))
                self.assertIn("Input does not match required columns", str(w[0].message))

    def test_query_frame_valid(self):
        with self.subTest('df'):
            pt.validate.query_frame(pd.DataFrame({'qid': [1, 2], 'query': ['cat', 'dog'], 'a': [1, 2]}), extra_columns=['a'])
        with self.subTest('cols'):
            pt.validate.query_frame(['qid', 'query', 'a'], extra_columns=['a'])
        with self.subTest('iterdict'):
            pt.validate.query_iter(pt.utils.peekable([{'qid': '1', 'query': 'cat', 'a': 1}]), extra_columns=['a'])

    def test_query_frame_invalid(self):
        with self.subTest('df'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.query_frame(pd.DataFrame({'query': ['cat', 'dog']}))
            self.assertIn('qid', cm.exception.modes[0].missing_columns)
        with self.subTest('cols'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.query_frame(['query'])
            self.assertIn('qid', cm.exception.modes[0].missing_columns)
        with self.subTest('iterdict'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.query_iter(pt.utils.peekable([{'query': 'cat'}]))
            self.assertIn('qid', cm.exception.modes[0].missing_columns)

    def test_document_frame_valid(self):
        with self.subTest('df'):
            pt.validate.document_frame(pd.DataFrame({'docno': [101, 102], 'text': ['hello', 'world']}))
        with self.subTest('cols'):
            pt.validate.document_frame(['docno', 'text'])
        with self.subTest('iterdict'):
            pt.validate.document_iter(pt.utils.peekable([{'docno': '101', 'text': 'hello'}]))

    def test_document_frame_invalid(self):
        with self.subTest('df'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.document_frame(pd.DataFrame({'text': ['hello', 'world']}))
            self.assertIn('docno', cm.exception.modes[0].missing_columns)
        with self.subTest('cols'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.document_frame(['text'])
            self.assertIn('docno', cm.exception.modes[0].missing_columns)
        with self.subTest('iterdict'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.document_iter(pt.utils.peekable([{'text': 'hello'}]))
            self.assertIn('docno', cm.exception.modes[0].missing_columns)

    def test_result_frame_valid(self):
        with self.subTest('df'):
            pt.validate.result_frame(pd.DataFrame({'qid': [1, 2], 'docno': [101, 102], 'score': [0.8, 0.6]}))
        with self.subTest('cols'):
            pt.validate.result_frame(['qid', 'docno', 'score'])
        with self.subTest('iterdict'):
            pt.validate.result_iter(pt.utils.peekable([{'qid': '1', 'docno': '101', 'score': 0.8}]))

    def test_result_frame_invalid(self):
        with self.subTest('df'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.result_frame(pd.DataFrame({'score': [0.8, 0.6]}))
            self.assertIn('qid', cm.exception.modes[0].missing_columns)
            self.assertIn('docno', cm.exception.modes[0].missing_columns)
        with self.subTest('cols'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.result_frame(['score'])
            self.assertIn('qid', cm.exception.modes[0].missing_columns)
            self.assertIn('docno', cm.exception.modes[0].missing_columns)
        with self.subTest('iterdict'):
            with self.assertRaises(pt.validate.InputValidationError) as cm:
                pt.validate.result_iter(pt.utils.peekable([{'score': 0.8}]))
            self.assertIn('qid', cm.exception.modes[0].missing_columns)
            self.assertIn('docno', cm.exception.modes[0].missing_columns)


if __name__ == '__main__':
    unittest.main()



import pyterrier as pt
from . import Transformer, Estimator
from .model import add_ranks
from typing import Sequence, Union, Tuple
import numpy as np
import pandas as pd

FeatureList = Union[Sequence[int], int]


class AblateFeatures(Transformer):
    
    def __init__(self, fids: FeatureList):
        self.fids = fids if isinstance(fids, list) else [fids]
        self.null = 0
        
    def transform(self, topics_and_res):
        pt.validate.result_frame(topics_and_res, extra_columns=["features"])
        if len(topics_and_res) == 0:
            return topics_and_res
        
        def _reset(row):
            fvalues = row["features"].copy() 
            for findex in self.fids:
                fvalues[findex] = self.null
            return fvalues
        
        topics_and_res = topics_and_res.copy()
        topics_and_res["features"] = topics_and_res.apply(_reset, axis=1)
        return topics_and_res

class KeepFeatures(Transformer):
    
    def __init__(self, fids : FeatureList):
        self.fids = fids if isinstance(fids, list) else [fids]
        self.null = 0
        
    def transform(self, topics_and_res):
        
        pt.validate.result_frame(topics_and_res, extra_columns=["features"])

        if len(topics_and_res) == 0:
            return topics_and_res
        
        topics_and_res = topics_and_res.copy()
        topics_and_res["features"] = topics_and_res.apply(lambda row: row["features"][self.fids], axis=1)
        return topics_and_res

class RegressionTransformer(Estimator):
    """
    This class simplifies the use of Scikit-learn's techniques for learning-to-rank.
    """
    def __init__(self, learner, *args, fit_kwargs={}, **kwargs):
        """
        Init method

        Args:
            LTR: The model which to use for learning-to-rank. Must have a fit() and predict() methods.
            fit_kwargs: A dictionary containing additional arguments that can be passed to LTR's fit() method.  
        """
        self.fit_kwargs = fit_kwargs
        super().__init__(*args, **kwargs)
        self.learner = learner
        self.num_f = None

    def fit(self, topics_and_results_Train, qrelsTrain, topics_and_results_Valid=None, qrelsValid=None):
        """
        Trains the model with the given topics.

        Args:
            topicsTrain(DataFrame): A dataframe with the topics to train the model
        """
        pt.validate.result_frame(topics_and_results_Train, extra_columns=["features"])
        if len(topics_and_results_Train) == 0:
            raise ValueError("No topics to fit to")
        train_DF = topics_and_results_Train.merge(qrelsTrain, on=['qid', 'docno'], how='left').fillna(0)
        kwargs = self.fit_kwargs
        self.learner.fit(np.stack(train_DF["features"].values), train_DF["label"].values, **kwargs)
        self.num_f = train_DF.iloc[0].features.shape[0]
        return self

    def transform(self, test_DF):
        """
        Predicts the scores for the given topics.

        Args:
            topicsTest(DataFrame): A dataframe with the test topics.
        """
        pt.validate.result_frame(test_DF, extra_columns=["features"])
        test_DF = test_DF.copy()

        # check for change in number of features
        found_numf = test_DF.iloc[0].features.shape[0]
        if self.num_f is not None:
            if found_numf != self.num_f:
                raise ValueError("Was trained using %d features, but found %d features for inference. This typically occurs because you have changed the pipeline between training and inference." % (self.num_f, found_numf))
        if hasattr(self.learner, 'feature_importances_'):
            if len(self.learner.feature_importances_) != found_numf:
                raise ValueError("Model %s expected %d features, but found %d features for inference. This typically occurs because you are reusing the learned model between different LTR pipelines." % (str(self.learner), len(self.learner.feature_importances_), found_numf))

        test_DF["score"] = self.learner.predict(np.stack(test_DF["features"].values))
        return add_ranks(test_DF)
    
    def transform_outputs(self, inp_cols):
        """
        Returns the output columns of the transformer.
        
        Args:
            inp_cols: The input columns to the transformer.
        """
        out = inp_cols.copy()
        if "score" not in out:
            out.append("score")
        if "rank" not in out:
            out.append("rank")
        return out

class LTRTransformer(RegressionTransformer):
    """
    This class simplifies the use of LightGBM and XGBoost for learning-to-rank.
    """

    def fit(self, topics_and_results_Train, qrelsTrain, topics_and_results_Valid, qrelsValid):
        """
        Trains the model with the given training and validation topics.

        Args:
            topics_and_results_Train(DataFrame): A dataframe with the topics and results to train the model
            qrelsTrain(DataFrame): A dataframe containing the qrels for the training topics
            topics_and_results_Valid(DataFrame): A dataframe with the topics and results for validation
            qrelsValid(DataFrame): A dataframe containing the qrels for the validation topics
            
        """

        def _count_by_topic(res : pd.DataFrame) -> Tuple[Sequence[int], pd.DataFrame]:
            # we must ensure res and count_series have the same ordering
            res = res.sort_values("qid")
            count_series = res.groupby(["qid"], sort=False)["docno"].count().to_numpy()
            return count_series, res

        if topics_and_results_Train is None or len(topics_and_results_Train) == 0:
            raise ValueError("No training results to fit to")
        if topics_and_results_Valid is None or len(topics_and_results_Valid) == 0:
            raise ValueError("No validation results to fit to")

        pt.validate.result_frame(topics_and_results_Train, extra_columns=["features"])
        pt.validate.result_frame(topics_and_results_Valid, extra_columns=["features"])

        tr_res = topics_and_results_Train.merge(qrelsTrain, on=['qid', 'docno'], how='left').fillna(0)
        va_res = topics_and_results_Valid.merge(qrelsValid, on=['qid', 'docno'], how='left').fillna(0)

        kwargs = self.fit_kwargs

        # this enforces a sort on tr_res and va_res that matches the counts
        counts_tr, tr_res = _count_by_topic(tr_res)
        counts_va, va_res = _count_by_topic(va_res)
        
        self.learner.fit(
            np.stack(tr_res["features"].values), tr_res["label"].values, 
            group=counts_tr, # we name group here for lightgbm compat. 
            eval_set=[(np.stack(va_res["features"].values), va_res["label"].values)],
            eval_group=[counts_va],
            **kwargs
        )
        self.num_f = tr_res.iloc[0].features.shape[0]

class FastRankEstimator(Estimator):
    """
    This class simplifies the use of FastRank's techniques for learning-to-rank.
    """
    def __init__(self, learner, *args, **kwargs):
        """
        Init method

        Args:
            learner: The model which to use for learning-to-rank. Must have a fit() and predict() methods.
        """
        super().__init__(*args, **kwargs)
        self.learner = learner
        self.model = None
        self.num_f = None

    def _make_dataset(self, test_DF, add_labels = False):
        
        from collections import defaultdict
        from itertools import count
        from fastrank import CDataset # type: ignore
        qid_map = defaultdict(count().__next__)
        features = np.stack(test_DF["features"].values).astype('float32')
        qids = test_DF["qid"].apply(lambda qid : qid_map[qid]).values
        if add_labels:
            y = test_DF["label"].values
        else:
            y = np.zeros(len(test_DF))
        dataset = CDataset.from_numpy(features, y, qids)
        return dataset

    def fit(self, topics_and_results_Train, qrelsTrain, topics_and_results_Valid=None, qrelsValid=None):

        pt.validate.result_frame(topics_and_results_Train, extra_columns=["features"])

        if topics_and_results_Train is None or len(topics_and_results_Train) == 0:
            raise ValueError("No training results to fit to")

        tr_res = topics_and_results_Train.merge(qrelsTrain, on=['qid', 'docno'], how='left').fillna(0)
        dataset = self._make_dataset(tr_res, add_labels=True)
        self.num_f = dataset.num_features()
        self.model = dataset.train_model(self.learner)

    def transform(self, topics_and_docs_Test):
        """
        Predicts the scores for the given topics.

        :param topics_and_docs_Test: A dataframe with the test topics.
        """
        pt.validate.result_frame(topics_and_docs_Test, extra_columns=["features"])
        if self.model is None:
            raise ValueError("fit() must be called first")
        test_DF = topics_and_docs_Test.copy()
        dataset = self._make_dataset(test_DF, add_labels=False)

        # check for change in number of features
        found_numf = dataset.num_features()
        if self.num_f is not None and found_numf != self.num_f:
            raise ValueError("FastRankEstimator expected %d features, but found %d features. This typically occurs because you have changed the pipeline between training and inference." % (self.num_f, found_numf))
        if hasattr(self.learner, 'feature_importances_'):
            if len(self.learner.feature_importances_) != found_numf:
                raise ValueError("FastRank model %s expected %d features, but found %d features" % (str(self.learner), len(self.learner.feature_importances_), found_numf))
        
        rtr = dataset.predict_scores(self.model)
        scores = [rtr[i] for i in range(len(rtr))]
        test_DF["score"] = scores
        return add_ranks(test_DF)
    
    def transform_outputs(self, inp_cols):
        """
        Returns the output columns of the transformer.
        
        Args:
            inp_cols: The input columns to the transformer.
        """
        out = inp_cols.copy()
        if "score" not in out:
            out.append("score")
        if "rank" not in out:
            out.append("rank")
        return out

def ablate_features(fids : FeatureList) -> Transformer:
    """
        Ablates features (sets feature value to 0) from a pipeline. This is useful for 
        performing feature ablation studies, whereby a feature is removed from the pipeline
        before learning. 

        :param fids: one or a list of integers corresponding to features indices to be removed
    """
    return AblateFeatures(fids)

def keep_features(fids : FeatureList) -> Transformer:
    """
        Reduces the features in a pipeline to only those mentioned. This is useful for 
        performing feature ablation studies, whereby only some features are kept 
        (and other removed) from a pipeline before learning occurs. 

        :param fids: one or a list of integers corresponding to the features indice to be kept
    """
    return KeepFeatures(fids)

def feature_to_score(fid : int) -> Transformer:
    """
        Applies a specified feature for ranking. Useful for evaluating which of a number of 
        pre-computed features are useful for ranking. 

        :param fid: a single feature id that should be kept
    """
    return pt.apply.doc_score(lambda row : row["features"][fid])

def apply_learned_model(learner, form : str = 'regression', **kwargs) -> Transformer:
    """
        Results in a transformer that can take in documents that have a "features" column,
        and pass that to the specified learner via its transform() function, to obtain the
        documents' "score" column. Learners should follow the sklearn's general pattern
        with a fit() method (
        c.f. an sklearn `Estimator <https://scikit-learn.org/stable/glossary.html#term-estimator>`_)
        and a `predict() <https://scikit-learn.org/stable/glossary.html#term-predict>`_ method.

        xgBoost and LightGBM are also supported through the use of `type='ltr'` kwarg.

        :param learner: an sklearn-compatible estimator
        :param form: either 'regression', 'ltr' or 'fastrank'        
    """
    if form == 'ltr':
        return LTRTransformer(learner, **kwargs)
    if form == 'fastrank':
        return FastRankEstimator(learner, **kwargs)
    return RegressionTransformer(learner, **kwargs)

def score_to_feature() -> Transformer:
    """
        Takes the document's "score" from the score attribute, and uses it as a single feature. 
        In particular, a feature union operator does not use any score of the documents in the
        candidate set as a ranking feaure. Using the resulting transformer within a feature-union
        means that an additional ranking feature is added to the "feature" column.

        Example::

            cands = pt.terrier.Retriever(index, wmodel="BM25")
            bm25f = pt.terrier.Retriever(index, wmodel="BM25F")
            pl2f = pt.terrier.Retriever(index, wmodel="PL2F")
            
            two_features = cands >> (bm25f  **  pl2f)
            three_features = cands >> (bm25f  **  pl2f ** pt.ltr.score_to_feature())  

    """
    return pt.apply.doc_features(lambda row : np.array(row["score"]))


import os
import inspect
import sys
from typing import Tuple, List, Callable, Set, Sequence, Union, Iterator, Iterable, Any, Optional
from contextlib import contextmanager
import platform
from functools import wraps
from importlib.metadata import EntryPoint
from importlib.metadata import entry_points as eps
import pyterrier as pt

def once() -> Callable:
    """
    Wraps a function that can only be called once. Subsequent calls will raise an error.
    """
    def _once(fn: Callable) -> Callable:
        called = False

        @wraps(fn)
        def _wrapper(*args, **kwargs):
            nonlocal called
            if called:
                raise ValueError(f"{fn.__name__} has already been run")
            # how to handle errors?
            res = fn(*args, **kwargs)
            called = True
            return res
        _wrapper.called = lambda: called  # type: ignore
        return _wrapper
    return _once


def entry_points(group: str) -> Tuple[EntryPoint, ...]:
    """
    A shim for Python<=3.X to support importlib.metadata.entry_points(group).
    Also ensures that no duplicates (by name) are returned.

    See <https://docs.python.org/3/library/importlib.metadata.html#entry-points> for more details.
    """
    orig_res: Sequence[EntryPoint]
    try:
        orig_res = tuple(eps(group=group)) # type: ignore # support EntryPoints.get() API on different python versions
    except TypeError:
        orig_res = tuple(eps().get(group, tuple())) # type: ignore # support EntryPoints.get() API on different python versions

    names = set()
    res = []
    for ep in orig_res:
        if ep.name not in names:
            res.append(ep)
            names.add(ep.name)

    return tuple(res)


def is_windows() -> bool:
    return platform.system() == 'Windows'


def noop(*args, **kwargs):
    pass


def set_tqdm(type=None):
    """
        Set the tqdm progress bar type that Pyterrier will use internally.
        Many PyTerrier transformations can be expensive to apply in some settings - users can
        view progress by using the verbose=True kwarg to many classes, such as terrier.Retriever.

        The `tqdm <https://tqdm.github.io/>`_ progress bar can be made prettier when using appropriately configured Jupyter notebook setups.
        We use this automatically when Google Colab is detected.

        Allowable options for type are:

         - `'tqdm'`: corresponds to the standard text progresss bar, ala `from tqdm import tqdm`.
         - `'notebook'`: corresponds to a notebook progress bar, ala `from tqdm.notebook import tqdm`
         - `'auto'`: allows tqdm to decide on the progress bar type, ala `from tqdm.auto import tqdm`. Note that this works fine on Google Colab, but not on Jupyter unless the `ipywidgets have been installed <https://ipywidgets.readthedocs.io/en/stable/user_install.html>`_.
    """
    import importlib
    if type is None:
        if 'google.colab' in sys.modules:
            type = 'notebook'
        elif _get_notebook() is not None and importlib.util.find_spec('ipywidgets') is not None:
            type = 'notebook'
        else:
            type = 'tqdm'

    if type == 'tqdm':
        from tqdm import tqdm as bartype
        pt.tqdm = bartype
    elif type == 'notebook':
        from tqdm.notebook import tqdm as bartype
        pt.tqdm = bartype
    elif type == 'auto':
        from tqdm.auto import tqdm as bartype
        pt.tqdm = bartype
    else:
        raise ValueError(f"Unknown tqdm type {type}")
    pt.tqdm.pandas()


def get_class_methods(cls) -> List[Tuple[str, Callable]]:
    """
    Returns methods defined directly by the provided class. This will ignore inherited methods unless they are
    overridden by this class.
    """
    all_attrs: Sequence[Tuple[str, Callable]] = inspect.getmembers(cls, predicate=inspect.isfunction)

    base_attrs : Set[str] = set()
    for base in cls.__bases__:
        base_attrs.update(name for name, _ in inspect.getmembers(base, predicate=inspect.isfunction))
    
    # Filter out methods that are in base classes and not overridden in the subclass
    class_methods : List[Tuple[str, Callable]] = []
    for name, func in all_attrs:
        if name not in base_attrs or func.__qualname__.split('.')[0] == cls.__name__:
            # bind classmethod and staticmethod functions to this class
            if any(isinstance(func, c) for c in [classmethod, staticmethod]):
                func = func.__get__(cls)
            class_methods.append((name, func))
    
    return class_methods


def pre_invocation_decorator(decorator):
    """
    Builds a decorator function that runs the decoraded code before running the wrapped function. It can run as a
    function @decorator, or a class @decorator. When used as a class decorator, it is applied to all functions defined
    by the class.
    """
    @wraps(decorator)
    def _decorator_wrapper(fn):
        if isinstance(fn, type): # wrapping a class
            for name, value in pt.utils.get_class_methods(fn):
                setattr(fn, name, _decorator_wrapper(value))
            return fn

        else: # wrapping a function
            @wraps(fn)
            def _wrapper(*args, **kwargs):
                decorator(fn)
                return fn(*args, **kwargs)
            return _wrapper
    return _decorator_wrapper


def byte_count_to_human_readable(byte_count: float) -> str:
    """Converts a byte count to a human-readable string."""
    units = ['B', 'KB', 'MB', 'GB', 'TB']
    while byte_count > 1024 and len(units) > 1:
        byte_count /= 1024
        units = units[1:]
    if units[0] == 'B':
        return f'{byte_count:.0f} {units[0]}'
    return f'{byte_count:.1f} {units[0]}'


@contextmanager
def temp_env(key: str, value: str):
    old_value = os.environ.get(key, None)
    try:
        os.environ[key] = value
        yield
    finally:
        if old_value is None:
            del os.environ[key]
        else:
            os.environ[key] = old_value


class GeneratorLen(object):
    def __init__(self, gen, length):
        self.gen = gen
        self.length = length

    def __len__(self): 
        return self.length

    def __iter__(self):
        return self.gen


_NO_BUFFER = object()


class PeekableIter:
    """An iterator that allows peeking at the next element."""
    def __init__(self, base: Union[Iterator, Iterable]):
        """Create a PeekableIter from an iterator or iterable."""
        self.base = iter(base)
        self._buffer = _NO_BUFFER

    def __getattr__(self, attr: str):
        return getattr(self.base, attr)

    def __next__(self):
        if self._buffer != _NO_BUFFER:
            n = self._buffer
            self._buffer = _NO_BUFFER
            return n
        return next(self.base)

    def __iter__(self):
        return self

    def peek(self) -> Any:
        """Return the next element without consuming it."""
        if self._buffer == _NO_BUFFER:
            self._buffer = next(self.base)
        return self._buffer


def peekable(it: Union[Iterator, Iterable]) -> PeekableIter:
    """Create a PeekableIter from an iterator or iterable."""
    return PeekableIter(it)

def _get_notebook() -> Optional[str]:
    """
    Returns the process name if running in a notebook, else None.
    """
    try:
        import IPython # type: ignore
    except Exception:
        return None

    # Try to get IPython and return None if not found.
    ipython = IPython.get_ipython()
    if not ipython:
        return None
    locals = IPython.get_ipython().user_ns

    if "__vsc_ipynb_file__" in locals:
        return locals["__vsc_ipynb_file__"]
    if "__session__" in locals:
        return locals["__session__"]
    return None


import pandas as pd
import unittest
import pyterrier as pt
import os
import shutil
import tempfile
import pyterrier.transformer as ptt;
from .base import TempDirTestCase

def normalize_term_weights(term_weights, digits=7):
    ret = ''
    for i in term_weights.split():
        if '^' in i:
            i = i.split('^')
            i = i[0] + '^' + i[1][:digits]
        ret += ' ' + i
    return ret.strip()

class TestRewrite(TempDirTestCase):

    def test_stash_results_noclear(self):
        inputDF = pt.new.ranked_documents([[1, 2], [2,0]], query=[["a", "a"], ["b", "b"]])
        savedDF = pt.rewrite.stash_results(clear=False)(inputDF)
        self.assertEqual(4, len(savedDF))
        self.assertIn("stashed_results_0", savedDF.columns)
        self.assertIn("qid", savedDF.columns)
        self.assertIn("docno", savedDF.columns)
        self.assertIn("score", savedDF.columns)
        self.assertIn("query", savedDF.columns)

        stasheddocs_q1 = savedDF.iloc[0]["stashed_results_0"]
        self.assertEqual(2, len(stasheddocs_q1))
        self.assertEqual(savedDF.iloc[0]["qid"], stasheddocs_q1[0]["qid"])

        stasheddocs_q2 = savedDF.iloc[3]["stashed_results_0"]
        self.assertEqual(2, len(stasheddocs_q2))
        self.assertEqual(savedDF.iloc[3]["qid"], stasheddocs_q2[0]["qid"])


    def test_stash_results(self):
        inputDF = pt.new.ranked_documents([[1, 2], [2,0]], query=[["a", "a"], ["b", "b"]])
        savedDF = pt.rewrite.stash_results()(inputDF)
        self.assertEqual(2, len(savedDF))
        self.assertIn("stashed_results_0", savedDF.columns)

        savedQueryDF = pt.apply.query(lambda row: row["query"] + " 1")(savedDF)
        self.assertIn("stashed_results_0", savedQueryDF.columns)
        self.assertIn("qid", savedQueryDF.columns)
        self.assertIn("query", savedQueryDF.columns)
        self.assertIn("query_0", savedQueryDF.columns)
        
        self.assertEqual(2, len(savedQueryDF))
        restoredDf = pt.rewrite.reset_results()(savedQueryDF)
        self.assertEqual(4, len(restoredDf))
        self.assertIn("qid", restoredDf.columns)
        self.assertIn("docno", restoredDf.columns)
        self.assertIn("score", restoredDf.columns)
        self.assertIn("query", restoredDf.columns)
    
    def test_stash_results_SDM(self):
        inputDF = pt.new.ranked_documents([[1, 2], [2,0]], query=[["a a", "a a"], ["b b", "b b"]])
        savedDF = pt.rewrite.stash_results()(inputDF)
        self.assertEqual(2, len(savedDF))
        self.assertIn("stashed_results_0", savedDF.columns)

        savedQueryDF = pt.rewrite.SDM()(savedDF)
        self.assertIn("stashed_results_0", savedQueryDF.columns)
        self.assertIn("qid", savedQueryDF.columns)
        self.assertIn("query", savedQueryDF.columns)
        self.assertIn("query_0", savedQueryDF.columns)
        
        self.assertEqual(2, len(savedQueryDF))
        restoredDf = pt.rewrite.reset_results()(savedQueryDF)
        self.assertEqual(4, len(restoredDf))
        self.assertIn("qid", restoredDf.columns)
        self.assertIn("docno", restoredDf.columns)
        self.assertIn("score", restoredDf.columns)
        self.assertIn("query", restoredDf.columns)

    def test_save_docs_CE(self):
        index = pt.get_dataset("vaswani").get_index()
        dph = pt.terrier.Retriever(index, wmodel="DPH")
        pipe = dph \
            >> pt.rewrite.stash_results() \
            >> pt.terrier.Retriever(index, wmodel="BM25") \
            >> pt.rewrite.Bo1QueryExpansion(index) \
            >> pt.rewrite.reset_results() \
            >> dph
        rtr1 = dph.search("chemical reactions")        
        rtr2 = pipe.search("chemical reactions")
        # Bo1 should be applied as a re-ranker, hence the
        # number of docs in rtr1 and rtr2 should be equal
        self.assertEqual(len(rtr1), len(rtr2))

        # check columns are passed through where we expect
        pipeP3 = dph \
            >> pt.rewrite.stash_results() \
            >> pt.terrier.Retriever(index, wmodel="BM25")
        res3 = pipeP3.search("chemical reactions")
        self.assertIn("stashed_results_0", res3.columns)
        pipeP4 = dph \
            >> pt.rewrite.stash_results() \
            >> pt.terrier.Retriever(index, wmodel="BM25") \
            >> pt.rewrite.Bo1QueryExpansion(index)
        res4 = pipeP3.search("chemical reactions")
        self.assertIn("stashed_results_0", res4.columns)
    
    def test_save_docs_QE(self):
        index = pt.get_dataset("vaswani").get_index()
        dph = pt.terrier.Retriever(index, wmodel="DPH")
        pipe = dph \
            >> pt.rewrite.stash_results(clear=False) \
            >> pt.rewrite.Bo1QueryExpansion(index) \
            >> pt.rewrite.reset_results() \
            >> dph
        rtr1 = dph.search("chemical reactions")        
        rtr2 = pipe.search("chemical reactions")
        # Bo1 should be applied as a re-ranker, hence the
        # number of docs in rtr1 and rtr2 should be equal
        self.assertEqual(len(rtr1), len(rtr2))

    def test_reset_with_docs(self):
        inputDF = pt.new.ranked_documents([[1, 2], [2,0]])
        inputDF["query"] = ["one #1", "one #1", "one two #1", "one two #1"]
        inputDF["query_0"] = ["one", "one", "one two", "one two"]
        outputDF = pt.rewrite.reset().transform(inputDF)
        self.assertEqual(len(inputDF), len(outputDF))
        self.assertTrue("query" in outputDF.columns)
        self.assertFalse("query_0" in outputDF.columns)
        self.assertTrue("docno" in outputDF.columns)
        self.assertTrue("score" in outputDF.columns)
        self.assertTrue("rank" in outputDF.columns)
        self.assertEqual(outputDF.iloc[0]["query"], "one")
        self.assertEqual(outputDF.iloc[1]["query"], "one")
        self.assertEqual(outputDF.iloc[2]["query"], "one two")
        self.assertEqual(outputDF.iloc[3]["query"], "one two")

    def test_sdm_freq(self):
        if not pt.terrier.check_version("5.3"):
            self.skipTest("Requires Terrier 5.3")
        self._sdm(True)

    def test_sdm(self):
        if not pt.terrier.check_version("5.3"):
            self.skipTest("Requires Terrier 5.3")
        self._sdm(False)

    def test_sdm_docs(self):
        docs = pt.new.ranked_documents([[1,1]], qid=["q1"], query=[["hello friend","hello friend"]])
        sdm = pt.rewrite.SDM()
        pipe = pt.Transformer.from_df(docs) >> sdm
        qids = pt.new.queries(["hello there"], qid=["q1"])  
        rtr = pipe(qids)
        self.assertIn("query", rtr.columns)
        self.assertIn("query_0", rtr.columns)
        self.assertIn("docno", rtr.columns)
        self.assertEqual(2, len(rtr))


    def _sdm(self, freq):
        dataset = pt.datasets.get_dataset("vaswani")
        indexer = pt.TRECCollectionIndexer(self.test_dir, blocks=True)
        indexref = indexer.index(dataset.get_corpus())
        if freq:
            sdm = pt.rewrite.SDM(prox_model="org.terrier.matching.models.Tf")
        else:
            sdm = pt.rewrite.SDM()
       
        queriesIn = pd.DataFrame([["1", "compact"], ["2", "compact memories"]], columns=["qid", "query"])
        queriesOut = sdm.transform(queriesIn)
        self.assertEqual(len(queriesOut), 2)
        self.assertIn("query", queriesOut.columns)
        self.assertIn("query_0", queriesOut.columns)
        self.assertEqual(queriesOut.iloc[0]["query"], "compact")

        # check for pushed query representation        
        self.assertTrue("query_0" in queriesOut.columns)
        self.assertEqual(queriesOut.iloc[0]["query_0"], "compact")

        query2 = queriesOut.iloc[1]["query"]
        self.assertTrue("#1" in query2)
        self.assertTrue("#uw8" in query2)
        self.assertTrue("#combine" in query2)
        self.assertEqual(queriesOut.iloc[1]["query_0"], "compact memories")
        
        br_normal = pt.terrier.Retriever(indexref)
        pipe = sdm >> br_normal

        #check that we can get a str()
        str(pipe)

        if freq:
            br_normal.controls["wmodel"] = "Tf"
        resTest_pipe = pipe.transform(queriesIn)


        # this BR should do the same thing as the pipe, but natively in Terrier
        br_sdm = pt.terrier.Retriever(indexref,
            controls = {"sd" :"on"}, 
            properties={"querying.processes" : "terrierql:TerrierQLParser,parsecontrols:TerrierQLToControls,"\
                    +"parseql:TerrierQLToMatchingQueryTerms,matchopql:MatchingOpQLParser,applypipeline:ApplyTermPipeline,"\
                    +"sd:DependenceModelPreProcess,localmatching:LocalManager$ApplyLocalMatching,qe:QueryExpansion,"\
                    +"labels:org.terrier.learning.LabelDecorator,filters:LocalManager$PostFilterProcess"})
        if freq:
            br_sdm.controls["wmodel"] = "Tf"
            br_sdm.controls["dependencemodel"] = "org.terrier.matching.models.Tf"

        resTest_native = br_sdm.transform(queriesIn)
 
        #print (resTest_pipe[resTest_pipe["qid"]=="2"])
        #print (resTest_native[resTest_native["qid"]=="2"])
        for index, row in resTest_pipe.iterrows():
            #print(index)
            #print(row["query"])
            #print(row)
            #print(resTest_native.iloc[index]) 
            self.assertEqual(row['qid'], resTest_native.iloc[index]["qid"])
            self.assertEqual(row['docno'], resTest_native.iloc[index]["docno"])
            # TODO I cannot get this test to pass with freq=False more precisely than 1dp
            #9.165638 in resTest_pipe vs 9.200683 in resTest_native
            self.assertAlmostEqual(row['score'], resTest_native.iloc[index]["score"], 1)

        t = dataset.get_topics().head(5)
        pipe_res = pipe.transform(t)
        #br_normal.saveResult(pipe_res, "/tmp/sdm.res", run_name="DPH")

        self.assertAlmostEqual(
            pt.Evaluate(pipe_res, dataset.get_qrels(), metrics=["map"])["map"], 
            pt.Evaluate(br_sdm.transform(t), dataset.get_qrels(), metrics=["map"])["map"], 
            places=4)

    def test_rm3_expansion_for_query_compact_on_tf_idf(self):
        # top-retrieval results of TF-IDF and BM25 below change, so the RM3 weights differ
        expected = 'applypipeline:off equip^0.037346367 ferrit^0.027371584 modul^0.027371584 suppli^0.037346367 design^0.056739070 microwav^0.027371584 anod^0.037346367 unit^0.037346367 compact^0.674414337 stabil^0.037346367'
        
        indexref = pt.datasets.get_dataset("vaswani").get_index()
        queriesIn = pd.DataFrame([["1", "compact"]], columns=["qid", "query"])

        qe = pt.rewrite.RM3(indexref)
        br = pt.terrier.Retriever(indexref, wmodel='TF_IDF')

        actual = qe.transform(br.transform(queriesIn))

        self.assertEqual(len(actual), 1)
        self.assertEqual(normalize_term_weights(expected), normalize_term_weights(actual.iloc[0]["query"]))

    def test_rm3_expansion_for_query_compact_on_bm25(self):
        # top-retrieval results of BM25 and TF-IDF above change, so the RM3 weights differ
        expected = 'applypipeline:off equip^0.040264644 ferrit^0.025508024 modul^0.025508024 suppli^0.040264644 design^0.051008239 microwav^0.025508024 anod^0.040264644 unit^0.040264644 compact^0.671144485 stabil^0.040264644'
        
        indexref = pt.datasets.get_dataset("vaswani").get_index()
        queriesIn = pd.DataFrame([["1", "compact"]], columns=["qid", "query"])

        qe = pt.rewrite.RM3(indexref)
        br = pt.terrier.Retriever(indexref, wmodel='BM25')

        actual = qe.transform(br.transform(queriesIn))

        self.assertEqual(len(actual), 1)
        self.assertEqual(normalize_term_weights(expected), normalize_term_weights(actual.iloc[0]["query"]))

    def test_kl_qe_expansion_for_query_compact_on_bm25(self):
        # just ensure that KLQueryExpansion results do not change
        expected = 'applypipeline:off compact^1.840895333 design^0.348370740 equip^0.000000000 purpos^0.000000000 instrument^0.000000000 ferrit^0.000000000 anod^0.000000000 aircraft^0.000000000 microwav^0.000000000 sideband^0.000000000'
        
        indexref = pt.datasets.get_dataset("vaswani").get_index()
        queriesIn = pd.DataFrame([["1", "compact"]], columns=["qid", "query"])

        qe = pt.rewrite.KLQueryExpansion(indexref)
        br = pt.terrier.Retriever(indexref, wmodel='BM25')

        actual = qe.transform(br.transform(queriesIn))

        self.assertEqual(len(actual), 1)
        self.assertEqual(expected, actual.iloc[0]["query"])

    def test_bo1_qe_expansion_for_query_compact_on_bm25(self):
        # just ensure that Bo1QueryExpansion results do not change
        expected = 'applypipeline:off compact^1.822309726 design^0.287992096 equip^0.000000000 purpos^0.000000000 instrument^0.000000000 ferrit^0.000000000 anod^0.000000000 aircraft^0.000000000 microwav^0.000000000 sideband^0.000000000'
        
        indexref = pt.datasets.get_dataset("vaswani").get_index()
        queriesIn = pd.DataFrame([["1", "compact"]], columns=["qid", "query"])

        qe = pt.rewrite.Bo1QueryExpansion(indexref)
        br = pt.terrier.Retriever(indexref, wmodel='BM25')

        actual = qe.transform(br.transform(queriesIn))

        self.assertEqual(len(actual), 1)
        self.assertEqual(expected, actual.iloc[0]["query"])

    def test_dfr_qe_expansion_for_query_compact_on_bm25(self):
        # just ensure that DFRQueryExpansion results do not change
        expected = 'applypipeline:off compact^1.822309726 design^0.287992096 equip^0.000000000 purpos^0.000000000 instrument^0.000000000 ferrit^0.000000000 anod^0.000000000 aircraft^0.000000000 microwav^0.000000000 sideband^0.000000000'
        
        indexref = pt.datasets.get_dataset("vaswani").get_index()
        queriesIn = pd.DataFrame([["1", "compact"]], columns=["qid", "query"])

        qe = pt.rewrite.DFRQueryExpansion(indexref)
        br = pt.terrier.Retriever(indexref, wmodel='BM25')

        actual = qe.transform(br.transform(queriesIn))

        self.assertEqual(len(actual), 1)
        self.assertEqual(expected, actual.iloc[0]["query"])

    def test_rm3_end_to_end(self):
        """An end-to-end test, contrasting the smaller tests (that fail faster) from above.
        """
        dataset = pt.datasets.get_dataset("vaswani")
        indexref = dataset.get_index()

        qe = pt.rewrite.RM3(indexref)
        br = pt.terrier.Retriever(indexref)

        queriesIn = pd.DataFrame([["1", "compact"]], columns=["qid", "query"])
        res = br.transform(queriesIn)

        queriesOut = qe.transform(res)
        self.assertEqual(len(queriesOut), 1)
        query = queriesOut.iloc[0]["query"]
        #self.assertTrue("compact^1.82230972" in query)
        self.assertTrue("applypipeline:off " in query)
        
        pipe = br >> qe >> br

        # lets go faster, we only need 18 topics. qid 16 had a tricky case
        t = dataset.get_topics().head(18)

        all_qe_res = pipe.transform(t)
        map_pipe = pt.Evaluate(all_qe_res, dataset.get_qrels(), metrics=["map"])["map"]

        br_qe = pt.terrier.Retriever(indexref, 
            controls={"qe":"on"},
            properties={"querying.processes" : "terrierql:TerrierQLParser,parsecontrols:TerrierQLToControls,"\
                    +"parseql:TerrierQLToMatchingQueryTerms,matchopql:MatchingOpQLParser,applypipeline:ApplyTermPipeline,"\
                    +"sd:DependenceModelPreProcess,localmatching:LocalManager$ApplyLocalMatching,qe:RM3,"\
                    +"labels:org.terrier.learning.LabelDecorator,filters:LocalManager$PostFilterProcess"})
        map_qe = pt.Evaluate(br_qe.transform(t), dataset.get_qrels(), metrics=["map"])["map"]

        self.assertAlmostEqual(map_qe, map_pipe, places=2)

    def test_scoring_rm3_qe(self):
        expected = 'applypipeline:off fox^0.600000024'
        input = pd.DataFrame([["q1", "fox", "d1", "all the fox were fox", 3], ["q1", "fox", "d2", "brown fox jumps", 2]], columns=["qid", "query", "docno", "body", "score"])
        scorer = pt.terrier.retriever.TextIndexProcessor(pt.rewrite.RM3, takes="docs", returns="queries")
        rtr = scorer(input)
        self.assertTrue("qid" in rtr.columns)
        self.assertTrue("query" in rtr.columns)
        self.assertTrue("docno" not in rtr.columns)
        self.assertTrue(expected, rtr.iloc[0]["query"])

    def test_linear_terrierql(self):
        pipe = pt.apply.query(lambda row: "az") >> pt.rewrite.linear(0.75, 0.25)
        self.assertEqual(pipe[1].weights[0], 0.75)
        self.assertEqual(pipe[1].weights[1], 0.25)
        dfIn = pt.new.queries(["a"])
        dfOut = pipe(dfIn)
        self.assertEqual(1, len(dfIn))
        self.assertEqual("az", dfOut.iloc[0]["query_0"])
        self.assertEqual("a", dfOut.iloc[0]["query_1"])
        self.assertEqual("(az)^0.750000 (a)^0.250000", dfOut.iloc[0]["query"])

    def test_linear_matchopql(self):
        pipe = pt.apply.query(lambda row: "az") >> pt.rewrite.linear(0.75, 0.25, format="matchopql")
        self.assertEqual(pipe[1].weights[0], 0.75)
        self.assertEqual(pipe[1].weights[1], 0.25)
        dfIn = pt.new.queries(["a"])
        dfOut = pipe(dfIn)
        self.assertEqual(1, len(dfIn))
        self.assertEqual("az", dfOut.iloc[0]["query_0"])
        self.assertEqual("a", dfOut.iloc[0]["query_1"])
        self.assertEqual("#combine:0:0.750000:1:0.250000(#combine(az) #combine(a))", dfOut.iloc[0]["query"])

    def test_tokenise(self):
        self.assertEqual('chemical reactions', pt.rewrite.tokenise().search("Chemical Reactions?").iloc[0].query)
        self.assertEqual('f r', pt.rewrite.tokenise().search("før").iloc[0].query)
        self.assertEqual('før', pt.rewrite.tokenise('utf').search("før").iloc[0].query)
        self.assertEqual('før', pt.rewrite.tokenise(pt.TerrierTokeniser.utf).search("før").iloc[0].query)
        self.assertEqual('før', pt.rewrite.tokenise('UTFTokeniser').search("før").iloc[0].query)
        self.assertEqual('før', pt.rewrite.tokenise('identity').search("før").iloc[0].query)
        self.assertEqual('#base64(ZsO4cg==)', pt.rewrite.tokenise('identity', matchop=True).search("før").iloc[0].query)
        self.assertEqual('a b', pt.rewrite.tokenise(lambda q : q.split('/')).search("a/b").iloc[0].query)

    def test_tokenise_hgf(self):
        try:
            from transformers import AutoTokenizer
        except:
            self.skipTest()
        tok = AutoTokenizer.from_pretrained("bert-base-uncased")
        query_toks = pt.rewrite.tokenise(tok.tokenize, matchop=True)
        self.assertEqual('a b', query_toks.search("a b").iloc[0].query)

    def test_qe(self):
        if not pt.terrier.check_version("5.3"):
            self.skipTest("Requires Terrier 5.3")
        dataset = pt.datasets.get_dataset("vaswani")
        indexref = dataset.get_index()
        index = pt.IndexFactory.of(indexref)

        qe3 = pt.rewrite.Bo1QueryExpansion(index)

        # lets go faster, we only need 18 topics. qid 16 had a tricky case
        t = dataset.get_topics().head(18)

        qrels = dataset.get_qrels()

        for qe in [qe3]:
            br = pt.terrier.Retriever(index)

            queriesIn = pd.DataFrame([["1", "compact"]], columns=["qid", "query"])
            res = br.transform(queriesIn)

            queriesOut = qe.transform(res)
            self.assertEqual(len(queriesOut), 1)
            self.assertTrue("query_0" in queriesOut.columns)
            self.assertEqual(queriesOut.iloc[0]["query_0"], "compact")
            query = queriesOut.iloc[0]["query"]
            self.assertTrue("compact^1.82230972" in query)
            self.assertTrue("applypipeline:off " in query)
            
            pipe = br >> qe >> br

            # check the pipe doesnt cause an error
            str(pipe)

            all_qe_res = pipe.transform(t)
            map_pipe = pt.Evaluate(all_qe_res, qrels, metrics=["map"])["map"]

            br_qe = pt.terrier.Retriever(indexref, controls={"qe":"on"})
            map_qe = pt.Evaluate(br_qe.transform(t), qrels, metrics=["map"])["map"]

            self.assertAlmostEqual(map_qe, map_pipe, places=4)

            pipe_opt = pipe.compile()
            self.assertEqual("2", pipe_opt[0].controls["end"]) # has fb_docs 3 been propagated back to the first retriever
            self.assertNotIn("end", pipe_opt[-1].controls) # ensure last retriever is unchanged
            pd.testing.assert_frame_equal(all_qe_res.head(5), pipe_opt(t).head(5))


@pt.testing.transformer_test_class
def test_sdm():
    return pt.rewrite.SDM()


@pt.testing.transformer_test_class
def test_rm3():
    index = pt.terrier.TerrierIndex.load(os.path.dirname(os.path.realpath(__file__)) + '/fixtures/index/')
    return pt.rewrite.RM3(index.index_ref())


if __name__ == "__main__":
    unittest.main()



import os
import tempfile 
import pyterrier as pt


def _wrap(text, width):
    return text
    #return '\\\n_'.join(textwrap.wrap(text, width=width))

def _get_text(row, name, width):
    value = row.get(name)
    if type(value) != list:
        return value
    if len(value) == 1:
        return value
    return '[' + _wrap(', '.join(value), width=width) + ']'

def fix_width(df, name, width=20):
    df[name] = df.apply(lambda row: _get_text(row, name, width), axis=1)
    return df


def setup():
    import os
    if not os.path.exists("_includes"):
        os.makedirs("_includes")

def dataset_include():
    df = pt.list_datasets()
    df = fix_width(df, "qrels")
    df = fix_width(df, "topics")
    df = fix_width(df, "corpus")
    df = df[["dataset", "corpus", "index", "topics", "qrels", "info_url"]]
    table = df.set_index('dataset').to_markdown(tablefmt="rst")

    with open("_includes/datasets-list-inc.rst", "wt") as f:
        f.write(table)

def experiment_includes():
    print("Regenerating experiments includes - to skip, use QUICK=1 make html")
    dataset = pt.get_dataset("vaswani")
    # vaswani dataset provides an index, topics and qrels

    # lets generate two BRs to compare
    try:
        indexref = dataset.get_index()
    except ValueError:
        # if data.terrier.org is down, build the index
        indexref = pt.IterDictIndexer(
                os.path.join(tempfile.gettempdir(), "vaswani_index")
            ).index(pt.get_dataset('vaswani').get_corpus_iter())

    tfidf = pt.terrier.Retriever(indexref, wmodel="TF_IDF")
    bm25 = pt.terrier.Retriever(indexref, wmodel="BM25")

    table = pt.Experiment(
        [tfidf, bm25],
        dataset.get_topics(),
        dataset.get_qrels(),
        eval_metrics=["map", "recip_rank"]
    ).to_markdown(tablefmt="rst")

    with open("_includes/experiment-basic.rst", "wt") as f:
        f.write(table)

    table = pt.Experiment(
        [tfidf, bm25],
        dataset.get_topics(),
        dataset.get_qrels(),
        eval_metrics=["map", "recip_rank"],
        names=["TF_IDF", "BM25"]
    ).to_markdown(tablefmt="rst")

    with open("_includes/experiment-names.rst", "wt") as f:
        f.write(table)

    table = pt.Experiment(
        [tfidf, bm25],
        dataset.get_topics(),
        dataset.get_qrels(),
        eval_metrics=["official"],
        names=["TF_IDF", "BM25"]
    ).to_markdown(tablefmt="rst")

    with open("_includes/experiment-official.rst", "wt") as f:
        f.write(table)

    table = pt.Experiment(
        [tfidf, bm25],
        dataset.get_topics(),
        dataset.get_qrels(),
        eval_metrics=["map", "recip_rank"],
        round={"map" : 4, "recip_rank" : 3},
        names=["TF_IDF", "BM25"]
    ).to_markdown(tablefmt="rst")

    with open("_includes/experiment-round.rst", "wt") as f:
        f.write(table)

    table = pt.Experiment(
        [tfidf, bm25],
        dataset.get_topics(),
        dataset.get_qrels(),
        eval_metrics=["map", "recip_rank"],
        names=["TF_IDF", "BM25"],
        baseline=0
    ).to_markdown(tablefmt="rst")
    with open("_includes/experiment-sig.rst", "wt") as f:
        f.write(table)

    table = pt.Experiment(
        [tfidf, bm25],
        dataset.get_topics(),
        dataset.get_qrels(),
        eval_metrics=["map"],
        names=["TF_IDF", "BM25"],
        baseline=0, correction='bonferroni'
    ).to_markdown(tablefmt="rst")
    with open("_includes/experiment-sig-corr.rst", "wt") as f:
        f.write(table)

    import statsmodels.stats.multitest
    import pandas as pd
    rows=[]
    for (shortcut, name), aliases in zip ( statsmodels.stats.multitest.multitest_methods_names.items(), statsmodels.stats.multitest._alias_list ):
        rows.append([aliases, name])
    table = pd.DataFrame(rows, columns=["Aliases", "Correction Method"]).to_markdown(tablefmt="rst")
    with open("_includes/experiment-corr-methods.rst", "wt") as f:
        f.write(table)

    
    table = pt.Experiment(
        [tfidf, bm25],
        dataset.get_topics(),
        dataset.get_qrels(),
        eval_metrics=["map", "recip_rank"],
        names=["TF_IDF", "BM25"],
        perquery=True
    ).head().to_markdown(tablefmt="rst")
    with open("_includes/experiment-perq.rst", "wt") as f:
        f.write(table)


def artifact_list_include():
    table = []
    for ep in pt.utils.entry_points('pyterrier.artifact'):
        table.append({
            'class' : ep.value.replace(':', '.'),
            'package': ep.value.replace(':', '.').split('.')[0],
            'type' : ep.name.split('.')[0],
            'format' : ep.name.split('.')[1],
        })
    table = sorted(table, key=lambda x: (x['package'] != 'pyterrier', 'index' not in x['type'], x['type'], x['format']))
    with open("_includes/artifact_list.rst", "wt") as f:
        f.write('''
.. list-table::
   :header-rows: 1

   * - Class (Package)
     - Type/Format
     - Artifacts on...
''')
        for rec in table:
            f.write('''
   * - :class:`~{class}` (``{package}``)
     - ``{type}/{format}``
     - `HuggingFace <https://huggingface.co/datasets?other=pyterrier-artifact.{type}.{format}>`__ | `Zenodo <https://zenodo.org/search?q=metadata.subjects.subject%3A%22pyterrier-artifact.{type}.{format}%22>`__
'''.format(**rec))



from typing import Callable, Any, Union, Optional, Iterable, List, Dict, Literal
import itertools
import more_itertools
import numpy.typing as npt
import pandas as pd
import pyterrier as pt


class DropColumnTransformer(pt.Transformer):
    """
    This transformer drops the provided column from the input.
    """
    def __init__(self, col: str):
        """
        Instantiates a DropColumnTransformer.

        Arguments:
            col: The column to drop
        """
        self.col = col

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """
        Drops the column from the input DataFrame.

        Arguments:
            inp: The input DataFrame

        Returns:
            The input DataFrame with the column dropped.
        """
        pt.validate.columns(inp, includes=[self.col])
        return inp.drop(columns=[self.col])

    def transform_iter(self, inp: pt.model.IterDict) -> pt.model.IterDict:
        """
        Drops the column from the input IterDict.

        Arguments:
            inp: The input IterDict

        Returns:
            The input with the column dropped.
        """
        inp = pt.utils.peekable(inp)
        pt.validate.columns_iter(inp, includes=[self.col])
        for rec in inp:
            new_rec = rec.copy()
            new_rec.pop(self.col, None) # None ensures no error if key doesn't exist
            yield new_rec

    def __repr__(self):
        return f"pt.apply.{self.col}(drop=True)"


class RenameColumnsTransformer(pt.Transformer):
    """
    This transformer renames the provided columns, akin to pandas.DataFrame.rename
    """
    def __init__(self, columns: Dict[str, str], *, errors: Literal['raise', 'ignore'] = 'raise'):
        """
        Arguments:
            columns: A dictionary mapping old column names to new column names
        """
        self.columns = columns
        self.errors = errors

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """
        Drops the column from the input DataFrame.

        Arguments:
            inp: The input DataFrame

        Returns:
            The input DataFrame with the column dropped.
        """
        if self.errors == 'raise':
            pt.validate.columns(inp, includes=list(self.columns.keys()))
        return inp.rename(columns=self.columns, errors=self.errors)

    def __repr__(self):
        return "pt.apply.rename()"


class ApplyByRowTransformer(pt.Transformer):
    """
    This transformer applies a function to each row in the input and assigns the result to a new column.
    """
    def __init__(self,
        col: str,
        fn: Callable[[Union[pt.model.IterDictRecord, pd.Series]], Any],
        *,
        batch_size: Optional[int] = None,
        required_columns: Optional[List[str]] = None,
        verbose: bool = False
    ):
        """
        Instantiates a ApplyByRowTransformer.

        Arguments:
            col: The column to assign the result of the function to
            fn: The function to apply to each row
            batch_size: The number of rows to process at once. If None, processes in one batch. This only applies
                when processing DataFrames.
            required_columns: A list of columns that must be present in the input DataFrame, or None to disable validation.
            verbose: Whether to display a progress bar when processing in batch mode.
        """
        self.col = col
        self.fn = fn
        self.batch_size = batch_size
        self.required_columns = required_columns
        self.verbose = verbose

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """
        Applies the function to each row in the input DataFrame and assigns the result to a new column.

        Arguments:
            inp: The input DataFrame

        Returns:
            The input DataFrame with the new values assign to ``col``
        """
        if self.required_columns is not None:
            pt.validate.columns(inp, includes=self.required_columns)

        if self.batch_size is None:
            return self._apply_df(inp)

        # batching
        iterator = pt.model.split_df(inp, batch_size=self.batch_size)
        if self.verbose:
            iterator = pt.tqdm(iterator, desc="pt.apply", unit='row')
        return pd.concat([self._apply_df(chunk_df) for chunk_df in iterator])

    def _apply_df(self, inp: pd.DataFrame) -> pd.DataFrame:
        new_vals = inp.apply(self.fn, axis=1, result_type='reduce')
        return inp.assign(**{self.col: new_vals})

    def transform_iter(self, inp: pt.model.IterDict) -> pt.model.IterDict:
        """
        Applies the function to each row in the input IterDict and assigns the result to a new column.

        Arguments:
            inp: The input IterDict

        Returns:
            The input IterDict with the new values assign to ``col``
        """
        for rec in inp:
            yield dict(rec, **{self.col: self.fn(rec)})

    def __repr__(self):
        return f"pt.apply.{self.col}()"


class ApplyForEachQuery(pt.Transformer):
    def __init__(self,
        fn: Callable[[pd.DataFrame], pd.DataFrame],
        *,
        add_ranks: bool = True,
        batch_size: Optional[int] = None,
        required_columns: Optional[List[str]] = None,
        verbose: bool = False
    ):
        """
        Instantiates a ApplyForEachQuery.

        Arguments:
            fn: Takes as input a DataFrame representing all results for a query and returns a transformed DataFrame
            add_ranks: Whether to calcualte and add ranks to the output for each query
            batch_size: The number of results per query to process at once. If None, processes in one batch per query.
            required_columns: A list of columns that must be present in the input DataFrame, or None to disable validation.
            verbose: Whether to display a progress bar
        """
        self.fn = fn
        self.add_ranks = add_ranks
        self.batch_size = batch_size
        self.required_columns = required_columns
        self.verbose = verbose

    def __repr__(self):
        return "pt.apply.by_query()"

    def transform(self, res: pd.DataFrame) -> pd.DataFrame:
        if self.required_columns is not None:
            pt.validate.columns(res, includes=self.required_columns)

        if len(res) == 0:
            return self.fn(res)

        it = res.groupby("qid")
        lastqid = None
        if self.verbose:
            it = pt.tqdm(it, unit='query')
        try:
            if self.batch_size is None:
                query_dfs = []
                for qid, group in it:
                    lastqid = qid
                    query_dfs.append(self.fn(group))
            else:
                # fn cannot be applied to more than batch_size rows at once
                # so we must split and reconstruct the output FOR EACH QUERY
                query_dfs = []
                for qid, group in it:
                    iterator = pt.model.split_df(group, batch_size=self.batch_size)
                    query_dfs.append( pd.concat([self.fn(chunk_df) for chunk_df in iterator]) )
        except Exception as a:
            raise Exception("Problem applying %r for qid %s" % (self.fn, lastqid)) from a # %r because its a function with bytes representation (mypy)

        if self.add_ranks:
            try:
                query_dfs = [pt.model.add_ranks(df, single_query=True) for df in query_dfs]
            except KeyError as ke:
                suffix = 'Try setting add_ranks=False'
                if len(query_dfs) > 0 and 'score' not in query_dfs[0].columns:
                    suffix = 'score column not present. Set add_ranks=False'
                raise ValueError("Cannot apply add_ranks in pt.apply.by_query - " + suffix) from ke
        rtr = pd.concat(query_dfs)
        return rtr


class ApplyIterForEachQuery(pt.Transformer):
    def __init__(self,
        fn: Callable[[pt.model.IterDict], pt.model.IterDict],
        *,
        batch_size=None,
        required_columns: Optional[List[str]] = None,
        verbose=False):
        """
        Instantiates a ApplyIterForEachQuery.

        Arguments:
            fn: Takes as input an IterDict of dictionaries representing all results for a query and returns a transformed IterDict
            batch_size: The number of results per query to process at once. If None, processes in one batch per query.
            required_columns: A list of columns that must be present in the input, or None to disable validation.
            verbose: Whether to display a progress bar
        """
        self.fn = fn
        self.batch_size = batch_size
        self.required_columns = required_columns
        self.verbose = verbose

    def __repr__(self):
        return "pt.apply.by_query()"

    def transform_iter(self, inp: pt.model.IterDict) -> pt.model.IterDict:
        if self.required_columns is not None:
            inp = pt.utils.peekable(inp)
            pt.validate.columns_iter(inp, includes=self.required_columns)

        if self.verbose:
            inp = pt.tqdm(inp, desc="pt.apply.by_query()")
        if self.batch_size is not None:
            for _, group in itertools.groupby(inp, key=lambda row: row['qid']):
                for batch in more_itertools.ichunked(group, self.batch_size):
                    yield from self.fn(batch)
        else:
            for _, group in itertools.groupby(inp, key=lambda row: row['qid']):
                yield from self.fn(group)


class ApplyDocumentScoringTransformer(pt.Transformer):
    """
        Implements a transformer that can apply a function to perform document scoring. The supplied function 
        should take as input one row, and return a float for the score of the document.
        
        Usually accessed using pt.apply.doc_score()::

            def _score_fn(row):
                return float(row["url".count("/")])
            
            pipe = pt.terrier.Retriever(index) >> pt.apply.doc_score(_score_fn)

        Can be used in batching manner, which is particularly useful for appling neural models. In this case,
        the scoring function receives a dataframe, rather than a single row::

            def _doclen(df):
                return df.text.str.len()
            
            pipe = pt.terrier.Retriever(index) >> pt.apply.doc_score(_doclen)

    """
    def __init__(self,
        fn: Union[
            Callable[[pd.Series], float],
            Callable[[pd.DataFrame], Iterable[float]],
        ],
        *,
        batch_size: Optional[int] = None,
        required_columns: Optional[List[str]] = None,
        verbose: bool = False,
    ):
        """
        Arguments:
            fn: Either takes a panda Series for a row (representing each document in the result set), and returns the
                new float doument score. Or, if batch_size is set, takes a DataFrame, and returns a sequence of floats
                representing scores for those documents.
            batch_size: How many documents to operate on at once. If None, operates row-wise.
            required_columns: A list of columns that must be present in the input DataFrame, or None to disable validation.
            verbose: Whether to display a progress bar
        """
        self.fn = fn
        self.batch_size = batch_size
        self.required_columns = required_columns
        self.verbose = verbose

    def __repr__(self):
        return "pt.apply.doc_score()"

    def _transform_rowwise(self, outputRes):
        if self.verbose:
            pt.tqdm.pandas(desc="pt.apply.doc_score", unit="d")
            outputRes["score"] = outputRes.progress_apply(self.fn, axis=1).astype('float64') # type: ignore
        else:
            outputRes["score"] = outputRes.apply(self.fn, axis=1).astype('float64')
        outputRes = pt.model.add_ranks(outputRes)
        return outputRes

    def _transform_batchwise(self, outputRes):
        outputRes["score"] = self.fn(outputRes)
        outputRes["score"] = outputRes["score"].astype('float64')
        return outputRes

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        if self.required_columns is not None:
            pt.validate.columns(inp, includes=self.required_columns)

        outputRes = inp.copy()
        if len(outputRes) == 0:
            outputRes["score"] = pd.Series(dtype='float64')
            return pt.model.add_ranks(outputRes)

        if self.batch_size is None:
            return self._transform_rowwise(outputRes)

        iterator = pt.model.split_df(outputRes, batch_size=self.batch_size)
        if self.verbose:
            iterator = pt.tqdm(iterator, desc="pt.apply", unit='row')
        rtr = pd.concat([self._transform_batchwise(chunk_df) for chunk_df in iterator])
        rtr = pt.model.add_ranks(rtr)
        return rtr


class ApplyDocFeatureTransformer(pt.Transformer):
    """
        Implements a transformer that can apply a function to perform feature scoring. The supplied function 
        should take as input one row, and return a numpy array for the features of the document.
        
        Usually accessed using pt.apply.doc_features()::

            def _feature_fn(row):
                return numpy.array([len(row["url"]), row["url"].count("/")])
            
            pipe = pt.terrier.Retriever(index) >> pt.apply.doc_features(_feature_fn) >> pt.LTRpipeline(xgBoost())
    """
    def __init__(self,
        fn: Callable[[Union[pd.Series, pt.model.IterDictRecord]], npt.NDArray],
        *,
        required_columns: Optional[List[str]] = None,
        verbose: bool = False
    ):
        """
        Arguments:
            fn: Takes as input a panda Series for a row representing that document, and returns a new numpy array representing the features of that document
            required_columns: A list of columns that must be present in the input DataFrame, or None to disable validation.
            verbose: Whether to display a progress bar
        """
        self.fn = fn
        self.required_columns = required_columns
        self.verbose = verbose

    def __repr__(self):
        return "pt.apply.doc_features()"
    
    def transform_output(self, cols: List[str]) -> List[str]:
        return cols + ["features"]

    def transform_iter(self, inp: pt.model.IterDict) -> pt.model.IterDict:
        if self.required_columns is not None:
            inp = pt.utils.peekable(inp)
            pt.validate.columns_iter(inp, includes=self.required_columns)

        # we assume that the function can take a dictionary as well as a pandas.Series. As long as [""] notation is used
        # to access fields, both should work
        if self.verbose:
            inp = pt.tqdm(inp, desc="pt.apply.doc_features")
        for row in inp:
            row["features"] = self.fn(row)
            yield row

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        if self.required_columns is not None:
            pt.validate.columns(inp, includes=self.required_columns)

        fn = self.fn
        outputRes = inp.copy()
        if self.verbose:
            pt.tqdm.pandas(desc="pt.apply.doc_features", unit="d")
            outputRes["features"] = outputRes.progress_apply(fn, axis=1) # type: ignore
        else:
            outputRes["features"] = outputRes.apply(fn, axis=1)
        return outputRes


class ApplyQueryTransformer(pt.Transformer):
    """
        Implements a query rewriting transformer by passing a function to perform the rewriting. The function should take
        as input one row, and return the string form of the new query.
        
        Usually accessed using pt.apply.query() passing it the function::

            def _rewriting_fn(row):
                return row["query"] + " extra words"
            
            pipe = pt.apply.query(_rewriting_fn) >> pt.terrier.Retriever(index)

        Similarly, a lambda function can also be used::

            pipe = pt.apply.query(lambda row: row["query"] + " extra words") >> pt.terrier.Retriever(index)

        In the resulting dataframe, the previous query for each row can be found in the query_0 column.

    """
    def __init__(self,
        fn: Callable[[Union[pd.Series, pt.model.IterDictRecord]], str],
        *,
        required_columns: Optional[List[str]] = None,
        verbose: bool = False
    ):
        """
        Arguments:
            fn: Takes as input a panda Series for a row representing a query, and returns the new string query 
            required_columns: A list of columns that must be present in the input DataFrame, or None to disable validation.
            verbose: Display a tqdm progress bar for this transformer
        """
        self.fn = fn
        self.required_columns = required_columns
        self.verbose = verbose

    def __repr__(self):
        return "pt.apply.query()"

    def transform_iter(self, inp: pt.model.IterDict) -> pt.model.IterDict:
        if self.required_columns is not None:
            inp = pt.utils.peekable(inp)
            pt.validate.columns_iter(inp, includes=self.required_columns)

        # we assume that the function can take a dictionary as well as a pandas.Series. As long as [""] notation is used
        # to access fields, both should work
        if self.verbose:
            inp = pt.tqdm(inp, desc="pt.apply.query")
        for row in inp:
            row = row.copy()
            if "query" in row:
                row = pt.model.push_queries_dict(row, inplace=True, keep_original=True)
            row["query"] = self.fn(row)
            yield row

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:  
        if self.required_columns is not None:
            pt.validate.columns(inp, includes=self.required_columns)

        if "query" in inp.columns:
            # we only push if a query already exists
            outputRes = pt.model.push_queries(inp.copy(), inplace=True, keep_original=True)
        else:
            outputRes = inp.copy()

        # handle empty DataFrame case
        if len(outputRes) == 0:
            outputRes["query"] = pd.Series(dtype='object')
            return outputRes
        
        try:
            if self.verbose:
                pt.tqdm.pandas(desc="pt.apply.query", unit="d")
                outputRes["query"] = outputRes.progress_apply(self.fn, axis=1) # type: ignore
            else:
                outputRes["query"] = outputRes.apply(self.fn, axis=1)
        except ValueError as ve:
            msg = str(ve)
            if "Columns must be same length as key" in msg:
                raise TypeError("Could not coerce return from pt.apply.query function into a list of strings. Check your function returns a string.") from ve
            else:
                raise ve
        return outputRes


class ApplyGenericTransformer(pt.Transformer):
    """
    Allows arbitrary pipelines components to be written as functions. The function should take as input
    a dataframe, and return a new dataframe. The function should abide by the main contracual obligations,
    e.g. updating then "rank" column.

    This class is normally accessed through pt.apply.generic()

    If you are scoring, query rewriting or calculating features, it is advised to use one of the other
    variants.

    Example::
        
        # this pipeline would remove all but the first two documents from all result sets
        lp = ApplyGenericTransformer(lambda res : res[res["rank"] < 2])

        pipe = pt.terrier.Retriever(index) >> lp

    """

    def __init__(self,
        fn: Callable[[pd.DataFrame], pd.DataFrame],
        *,
        batch_size: Optional[int] = None,
        required_columns: Optional[List[str]] = None,
        verbose: bool = False,
    ):
        """
        Arguments:
            fn: Takes as input a panda DataFrame, and returns a new Pandas DataFrame
            batch_size: The number of rows to process at once. If None, processes in one batch.
            verbose: When in batch model, display a tqdm progress bar
        """
        self.fn = fn
        self.batch_size = batch_size
        self.required_columns = required_columns
        self.verbose = verbose

    def __repr__(self):
        return "pt.apply.generic()"

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        if self.required_columns is not None:
            pt.validate.columns(inp, includes=self.required_columns)

        # no batching
        if self.batch_size is None:
            return self.fn(inp)

        # batching
        iterator = pt.model.split_df(inp, batch_size=self.batch_size)
        if self.verbose:
            iterator = pt.tqdm(iterator, desc="pt.apply", unit='row')
        rtr = pd.concat([self.fn(chunk_df) for chunk_df in iterator])
        return rtr


class ApplyGenericIterTransformer(pt.Transformer): 
    """

    As per ApplyGenericTransformer, but implements transform_iter(), not transform(). The supplied function
    is assumed to take Iterable[dict] and return Iterator[dict]

    This class is normally accessed through pt.apply.generic()

    If you are scoring, query rewriting or calculating features, it is advised to use one of the other
    variants.

    """
    def __init__(self,
        fn: Callable[[pt.model.IterDict], pt.model.IterDict],
        *,
        batch_size: Optional[int] = None,
        required_columns: Optional[List[str]] = None,
    ):
        """
        Arguments:
            fn: Takes as input a panda DataFrame, and returns a new Pandas DataFrame
            batch_size: The number of rows to process at once. If None, processes in one batch.
            required_columns: A list of columns that must be present in the input, or None to disable validation.
        """
        self.fn = fn
        self.batch_size = batch_size
        self.required_columns = required_columns

    def __repr__(self):
        return "pt.apply.generic(, iter=True)"

    def transform_iter(self, inp: pt.model.IterDict) -> pt.model.IterDict:
        if self.required_columns is not None:
            inp = pt.utils.peekable(inp)
            pt.validate.columns_iter(inp, includes=self.required_columns)

        if self.batch_size is None:
            # no batching
            yield from self.fn(inp)
        else:
            for batch in more_itertools.ichunked(inp, self.batch_size):
                yield from self.fn(batch)


class ApplyIndexer(pt.Indexer):
    """
    Allows arbitrary indexer pipelines components to be written as functions.
    """
    
    def __init__(self, fn: Callable[[pt.model.IterDict], Any], required_columns : Optional[List[str]] = None):
        self.fn = fn
        self.required_columns = required_columns

    def index(self, iter_dict):
        return self.fn(iter_dict)
    
    def index_inputs(self) -> Optional[List[List[str]]]:
        if self.required_columns is not None:
            return [self.required_columns]
        return self.required_columns



import pyterrier as pt

import unittest
import os
import shutil
import tempfile

from pyterrier.terrier.stemmer import TerrierStemmer
from .base import TempDirTestCase

class TestTRECIndexer(TempDirTestCase):

    def test_TREC_indexing_pbar(self):
        print("Writing index to " + self.test_dir)
        indexer = pt.TRECCollectionIndexer(self.test_dir, verbose=True)
        indexRef = indexer.index(pt.io.find_files(self.here + "/fixtures/vaswani_npl/corpus/"))
        self.assertIsNotNone(indexRef)
        index = pt.IndexFactory.of(indexRef)
        self.assertEqual(11429, index.getCollectionStatistics().getNumberOfDocuments())
        self.assertTrue(os.path.isfile(self.test_dir + '/data.direct.bf'))

    def test_TREC_indexing(self):
        print("Writing index to " + self.test_dir)
        indexer = pt.TRECCollectionIndexer(self.test_dir)
        indexRef = indexer.index(pt.io.find_files(self.here + "/fixtures/vaswani_npl/corpus/"))
        self.assertIsNotNone(indexRef)
        index = pt.IndexFactory.of(indexRef)
        self.assertEqual(11429, index.getCollectionStatistics().getNumberOfDocuments())
        self.assertTrue(os.path.isfile(self.test_dir + '/data.direct.bf'))

    def test_TREC_indexing(self):
        print("Writing index to " + self.test_dir)
        indexer = pt.TRECCollectionIndexer(self.test_dir, stemmer=TerrierStemmer.weakporter)
        indexer.index(pt.io.find_files(self.here + "/fixtures/vaswani_npl/corpus/"))
        index = pt.IndexFactory.of(self.test_dir)
        self.assertEqual(11429, index.getCollectionStatistics().getNumberOfDocuments())
        from jnius import cast as cst
        pindex = cst("org.terrier.structures.PropertiesIndex", index)
        self.assertTrue("WeakPorter" in pindex.getIndexProperty("termpipelines", "notfound"))

    def test_TREC_indexing_revmeta(self):
        print("Writing index to " + self.test_dir)
        indexer = pt.TRECCollectionIndexer(self.test_dir, meta_reverse=['docno'])
        indexRef = indexer.index(pt.io.find_files(self.here + "/fixtures/vaswani_npl/corpus/"))
        self.assertIsNotNone(indexRef)
        index = pt.IndexFactory.of(indexRef)
        self.assertEqual(11429, index.getCollectionStatistics().getNumberOfDocuments())
        self.assertTrue(os.path.isfile(self.test_dir + '/data.direct.bf'))
        meta = index.getMetaIndex()
        self.assertTrue('docno' in meta.getReverseKeys())
        self.assertEqual( meta.getDocument("docno", meta.getItem("docno", 2)), 2) 

    def test_TREC_indexing_text(self):
        print("Writing index to " + self.test_dir)
        indexer = pt.TRECCollectionIndexer(
            self.test_dir,
            meta= {'docno' : 26, 'body' : 2048},
            # The tags from which to save the text. ELSE is special tag name, which means anything not consumed by other tags.
            meta_tags = {'body' : 'ELSE'}
            )
        indexRef = indexer.index(pt.io.find_files(self.here + "/fixtures/vaswani_npl/corpus/"))
        self.assertIsNotNone(indexRef)
        index = pt.IndexFactory.of(indexRef)
        self.assertTrue("body" in index.getMetaIndex().getKeys())
        self.assertTrue("compact memories have flexible capacities" in index.getMetaIndex().getItem("body", 0) )
        self.assertEqual(11429, index.getCollectionStatistics().getNumberOfDocuments())
        self.assertTrue(os.path.isfile(self.test_dir + '/data.direct.bf'))

    def test_TREC_indexing_singlepass(self):
        print("Writing index to " + self.test_dir)
        indexer = pt.TRECCollectionIndexer(self.test_dir, type=pt.IndexingType.SINGLEPASS)
        indexRef = indexer.index(pt.io.find_files(self.here + "/fixtures/vaswani_npl/corpus/"))
        self.assertIsNotNone(indexRef)
        index = pt.IndexFactory.of(indexRef)
        self.assertEqual(11429, index.getCollectionStatistics().getNumberOfDocuments())
        self.assertFalse(os.path.isfile(self.test_dir + '/data.direct.bf'))

    def test_TREC_indexing_memory(self):
        from jnius import JavaException
        try:
            print("Writing index to " + self.test_dir)
            indexer = pt.TRECCollectionIndexer(self.test_dir, type=pt.IndexingType.MEMORY)
            indexRef = indexer.index(pt.io.find_files(self.here + "/fixtures/vaswani_npl/corpus/"))
            self.assertIsNotNone(indexRef)
            index = pt.IndexFactory.of(indexRef)
            self.assertEqual(11429, index.getCollectionStatistics().getNumberOfDocuments())
        except JavaException as ja:
            print(ja.stacktrace)
            raise ja

    def test_TREC_indexing_bad_files_type(self):
        print("Writing index to " + self.test_dir)
        indexer = pt.TRECCollectionIndexer(self.test_dir)
        with self.assertRaises(ValueError):
            indexRef = indexer.index(5)
    
        indexer = pt.TRECCollectionIndexer(self.test_dir)
        with self.assertRaises(ValueError):
            indexRef = indexer.index(pt.get_dataset("vaswani").get_corpus_iter())
    


import pandas as pd
import pyterrier as pt
import os
import unittest
from .base import BaseTestCase
import warnings

class TestTerrierWrappers(BaseTestCase):

    def test_stemming(self):
        stemmer = pt.TerrierStemmer.porter
        TESTS = {
            "abandoned": "abandon",
    		"abandon": "abandon",
	    	"abergavenny": "abergavenni"
        }
        for i,j in TESTS.items():
            self.assertEqual(j, stemmer.stem(i))

        stemmer = pt.TerrierStemmer.none
        for i,j in TESTS.items():
            self.assertEqual(i, stemmer.stem(i))

        stemmer = pt.TerrierStemmer.portugese
        for i,j in TESTS.items():
            self.assertTrue(len(stemmer.stem(i)) > 0)

if __name__ == "__main__":
    unittest.main()



from .transformer import Transformer, Estimator, get_transformer, SupportsFuseFeatureUnion, SupportsFuseRankCutoff, SupportsFuseRight, SupportsFuseLeft
from .model import add_ranks
from collections import deque
from warnings import warn
from typing import Optional, Iterable, Tuple, Iterator, List
from itertools import chain
import pandas as pd
import pyterrier as pt

class NAryTransformerBase(Transformer):
    """
        A base class for all operator transformers that can combine the input of 2 or more transformers. 
    """
    def __init__(self, *transformers: Transformer):
        assert len(transformers) > 0
        # Flatten out multiple layers of the same NAryTransformer into one
        transformers = _flatten(transformers, type(self))
        # Coerce datatypes
        self._transformers = tuple(get_transformer(x, stacklevel=6) for x in transformers)

    def __getitem__(self, number) -> Transformer:
        """
            Allows access to the ith transformer.
        """
        return self._transformers[number]

    def __len__(self) -> int:
        """
            Returns the number of transformers in the operator.
        """
        return len(self._transformers)

    def __iter__(self) -> Iterator[Transformer]:
        """
            Returns an iterator over the transformers in this pipeline.
        """
        return iter(self._transformers)

def _flatten(transformers: Iterable[Transformer], cls: type) -> Tuple[Transformer, ...]:
    return tuple(chain.from_iterable(
        (t._transformers if isinstance(t, cls) else [t]) # type: ignore[attr-defined]
        for t in transformers
    ))

class SetUnion(Transformer):
    """      
        This operator makes a retrieval set that includes documents that occur in the union (either) of both retrieval sets. 
        For instance, let left and right be pandas dataframes, both with the columns = [qid, query, docno, score], 
        left = [1, "text1", doc1, 0.42] and right = [1, "text1", doc2, 0.24]. 
        Then, left | right will be a dataframe with only the columns [qid, query, docno] and two rows = [[1, "text1", doc1], [1, "text1", doc2]].
                
        In case of duplicated both containing (qid, docno), only the first occurrence will be used.
    """
    def __init__(self, left: Transformer, right: Transformer):
        self.left = left
        self.right = right

    schematic = {'label': 'SetUnion |', 'inner_pipelines_mode': 'linked'}

    def transform(self, topics):
        res1 = self.left.transform(topics)
        res2 = self.right.transform(topics)
        import pandas as pd
        assert isinstance(res1, pd.DataFrame)
        assert isinstance(res2, pd.DataFrame)
        rtr = pd.concat([res1, res2])
        
        on_cols = ["qid", "docno"]     
        rtr = rtr.drop_duplicates(subset=on_cols)
        rtr = rtr.sort_values(by=on_cols)
        rtr.drop(columns=["score", "rank"], inplace=True, errors='ignore')
        return rtr

class SetIntersection(Transformer):
    """
        This operator makes a retrieval set that only includes documents that occur in the intersection of both retrieval sets. 
        For instance, let left and right be pandas dataframes, both with the columns = [qid, query, docno, score], 
        left = [[1, "text1", doc1, 0.42]] (one row) and right = [[1, "text1", doc1, 0.24],[1, "text1", doc2, 0.24]] (two rows).
        Then, left & right will be a dataframe with only the columns [qid, query, docno] and one single row = [[1, "text1", doc1]].
                
        For columns other than (qid, docno), only the left value will be used.
    """
    def __init__(self, left: Transformer, right: Transformer):
        self.left = left
        self.right = right

    schematic = {'label': 'SetIntersection &', 'inner_pipelines_mode': 'linked'}

    def transform(self, topics):
        res1 = self.left.transform(topics)
        res2 = self.right.transform(topics)  
        
        on_cols = ["qid", "docno"]
        rtr = res1.merge(res2, on=on_cols, suffixes=('','_y'))
        rtr.drop(columns=["score", "rank", "score_y", "rank_y", "query_y"], inplace=True, errors='ignore')
        for col in rtr.columns:
            if '_y' not in col:
                continue
            new_name = col.replace('_y', '')
            if new_name in rtr.columns:
                # duplicated column, drop
                rtr.drop(columns=[col], inplace=True)
                continue
            # column only from RHS, keep, but rename by removing '_y' suffix
            rtr.rename(columns={col:new_name}, inplace=True)

        return rtr

class Sum(Transformer):
    """
        Adds the scores of documents from two different retrieval transformers.
        Documents not present in one transformer are given a score of 0.
    """
    def __init__(self, left: Transformer, right: Transformer):
        self.left = left
        self.right = right

    schematic = {'label': 'Sum +', 'inner_pipelines_mode': 'linked'}

    def transform(self, topics_and_res):
        res1 = self.left.transform(topics_and_res)
        res2 = self.right.transform(topics_and_res)
        both_cols = set(res1.columns) & set(res2.columns)
        both_cols.remove("qid")
        both_cols.remove("docno")
        merged = res1.merge(res2, on=["qid", "docno"], suffixes=[None, "_r"], how='outer')
        merged["score"] = merged["score"].fillna(0) + merged["score_r"].fillna(0)
        merged = merged.drop(columns=["%s_r" % col for col in both_cols])
        merged = add_ranks(merged)
        return merged

class Concatenate(Transformer):
    epsilon = 0.0001

    def __init__(self, left: Transformer, right: Transformer):
        self.left = left
        self.right = right

    def transform(self, topics_and_res):
        import pandas as pd
        # take the first set as the top of the ranking
        res1 = self.left.transform(topics_and_res)
        # identify the lowest score for each query
        last_scores = res1[['qid', 'score']].groupby('qid').min().rename(columns={"score" : "_lastscore"})

        # the right hand side will provide the rest of the ranking        
        res2 = self.right.transform(topics_and_res)

        
        intersection = pd.merge(res1[["qid", "docno"]], res2[["qid", "docno"]].reset_index())
        remainder = res2.drop(intersection["index"])

        # we will append documents from remainder to res1
        # but we need to offset the score from each remaining document based on the last score in res1
        # explanation: remainder["score"] - remainder["_firstscore"] - self.epsilon ensures that the
        # first document in remainder has a score of -epsilon; we then add the score of the last document
        # from res1
        first_scores = remainder[['qid', 'score']].groupby('qid').max().rename(columns={"score" : "_firstscore"})

        remainder = remainder.merge(last_scores, on=["qid"]).merge(first_scores, on=["qid"])
        remainder["score"] = remainder["score"] - remainder["_firstscore"] + remainder["_lastscore"] - self.epsilon
        remainder = remainder.drop(columns=["_lastscore",  "_firstscore"])

        # now bring together and re-sort
        # this sort should match trec_eval
        rtr = pd.concat([res1, remainder]).sort_values(by=["qid", "score", "docno"], ascending=[True, False, True]) 

        # recompute the ranks
        rtr = add_ranks(rtr)
        return rtr

class ScalarProduct(Transformer):
    """
        Multiplies the retrieval score by a scalar
    """
    def __init__(self, scalar: float):
        self.scalar = scalar

    def transform(self, inp):
        out = inp.assign(score=inp["score"] * self.scalar)
        if self.scalar < 0:
            out = add_ranks(out)
        return out
    
    def schematic(self, *, input_columns = None): 
        return {'label': f'* {self.scalar}'}

    def __repr__(self):
        return f'ScalarProduct({self.scalar!r})'

class RankCutoff(Transformer):
    """
        Filters the input by rank<k for each query in the input
    """
    def __init__(self, k: int = 1000):
        self.k = k

    def transform(self, inp):
        assert 'rank' in inp.columns, "require rank to be present in the result set"
        res = inp[inp["rank"] < self.k]
        res = res.reset_index(drop=True)
        return res

    def __repr__(self):
        return f'RankCutoff({self.k!r})'
    
    def schematic(self, *, input_columns = None): 
        return {'label': f'% {self.k}'}

    def fuse_left(self, left: Transformer) -> Optional[Transformer]:
        # If the preceding component supports a native rank cutoff (via fuse_rank_cutoff), apply it.
        if isinstance(left, SupportsFuseRankCutoff):
            return left.fuse_rank_cutoff(self.k)
        return None

class FeatureUnion(NAryTransformerBase):
    """
        Implements the feature union operator.

        Example::
            cands = pt.terrier.Retriever(index wmodel="BM25")
            pl2f = pt.terrier.Retriever(index wmodel="PL2F")
            bm25f = pt.terrier.Retriever(index wmodel="BM25F")
            pipe = cands >> (pl2f ** bm25f)
    """
    schematic = {'inner_pipelines_mode': 'linked', 'label': 'FeatureUnion **'}

    def transform_inputs(self) -> Optional[List[List[str]]]:
        this_minimum : List[Optional[List[List[str]]]] = [[['qid', 'docno']]]
        return pt.inspect._minimal_inputs(this_minimum + [ 
            pt.inspect.transformer_inputs(t) for t in self._transformers
        ])

    def transform(self, inputRes):
        pt.validate.result_frame(inputRes)

        num_results = len(inputRes)
        import numpy as np

        # a parent could be a feature union, but it still passes the inputRes directly, so inputRes should never have a features column
        if "features" in inputRes.columns:
            raise ValueError("FeatureUnion operates as a re-ranker. They can be nested, but input "
                "should not contain a features column; found columns were %s" %  str(inputRes.columns))
        
        all_results = []

        for i, m in enumerate(self._transformers):
            # IMPORTANT this .copy() is important, in case an operand transformer changes inputRes
            results = m.transform(inputRes.copy())
            if len(results) == 0 and num_results != 0:
                raise ValueError("Got no results from %s, expected %d" % (repr(m), num_results) )
            assert "features_x" not in results.columns 
            assert "features_y" not in results.columns
            all_results.append( results )

    
        for i, (m, res) in enumerate(zip(self._transformers, all_results)):
            # IMPORTANT: dont do this BEFORE calling subsequent feature unions
            if "features" not in res.columns:
                if "score" not in res.columns:
                    raise ValueError("Results from %s did not include either score or features columns, found columns were %s" % (repr(m), str(res.columns)) )

                if len(res) != num_results:
                    warn(
                        "Got number of results different expected from %s, expected %d received %d, feature scores for any "
                        "missing documents be 0, extraneous documents will be removed" % (repr(m), num_results, len(res)))
                    all_results[i] = res = inputRes[["qid", "docno"]].merge(res, on=["qid", "docno"], how="left")
                    res["score"] = res["score"].fillna(value=0)

                if len(res) == 0:
                    res["features"] = pd.Series([], dtype='float64')
                else:
                    res["features"] = res.apply(lambda row : np.array([row["score"]]), axis=1)
                res.drop(columns=["score"], inplace=True)
            assert "features" in res.columns
            #print("%d got %d features from operand %d" % ( id(self) ,   len(results.iloc[0]["features"]), i))

        def _concat_features(row):
            assert isinstance(row["features_x"], np.ndarray)
            assert isinstance(row["features_y"], np.ndarray)
            
            left_features = row["features_x"]
            right_features = row["features_y"]
            return np.concatenate((left_features, right_features))
        
        def _reduce_fn(left, right):
            import pandas as pd
            both_cols = set(left.columns) & set(right.columns)
            both_cols.remove("qid")
            both_cols.remove("docno")
            both_cols.remove("features")
            rtr = pd.merge(left, right, on=["qid", "docno"])
            rtr["features"] = rtr.apply(_concat_features, axis=1, result_type='reduce')
            rtr.rename(columns={"%s_x" % col : col for col in both_cols}, inplace=True)
            rtr.drop(columns=["features_x", "features_y"] + ["%s_y" % col for col in both_cols], inplace=True)
            return rtr
        
        from functools import reduce
        final_DF = reduce(_reduce_fn, all_results)

        # final_DF should have the features column
        assert "features" in final_DF.columns

        # we used .copy() earlier, inputRes should still have no features column
        assert "features" not in inputRes.columns

        # final merge - this brings us the score attribute from any previous transformer
        both_cols = set(inputRes.columns) & set(final_DF.columns)
        both_cols.remove("qid")
        both_cols.remove("docno")
        final_DF = inputRes.merge(final_DF, on=["qid", "docno"])
        final_DF.rename(columns={"%s_x" % col : col for col in both_cols}, inplace=True)
        final_DF.drop(columns=["%s_y" % col for col in both_cols], inplace=True)
        # remove the duplicated columns
        #final_DF = final_DF.loc[:,~final_DF.columns.duplicated()]
        assert "features_x" not in final_DF.columns 
        assert "features_y" not in final_DF.columns 
        return final_DF

    def compile(self) -> Transformer:
        """
            Returns a new transformer that fuses feature unions where possible.
        """
        out : deque = deque()
        inp = deque([t.compile() for t in self._transformers])
        while inp:
            right = inp.popleft()
            if out and isinstance(out[-1], SupportsFuseFeatureUnion) and (fused := out[-1].fuse_feature_union(right, is_left=True)) is not None:
                out.pop()
                inp.appendleft(fused)
            elif out and isinstance(right, SupportsFuseFeatureUnion) and (fused := right.fuse_feature_union(out[-1], is_left=False)) is not None:
                out.pop()
                inp.appendleft(fused)
            else:
                out.append(right)
        if len(out) == 1:
            return out[0]
        return FeatureUnion(*out)

    def __repr__(self):
        return '(' + ' ** '.join([str(t) for t in self._transformers]) + ')'

class Compose(NAryTransformerBase):
    """ 
        This class allows pipeline components to be chained together using the "then" operator.

        :Example:

        >>> comp = ComposedPipeline([ DPH_br, ApplyGenericTransformer(lambda res : res[res["rank"] < 2])])
        >>> # OR
        >>> # we can even use lambdas as transformers
        >>> comp = ComposedPipeline([DPH_br, lambda res : res[res["rank"] < 2]])
        >>> #this is equivelent
        >>> #comp = DPH_br >> lambda res : res[res["rank"] < 2]]
    """
    name = "Compose"

    def index(self, iter : pt.model.IterDict, batch_size=None):
        """
        This methods implements indexing pipelines. It is responsible for calling the transform_iter() method of its 
        constituent transformers (except the last one) on batches of records, and the index() method on the last transformer.
        """
        from more_itertools import chunked
        
        prev_transformer = Compose(*self._transformers[0:-1])
        last_transformer = self._transformers[-1]

        # guess a good batch size from the batch_size of individual components earlier in the pipeline
        if batch_size is None:
            batch_size = 100 # default to 100 as a reasonable minimum (and fallback if no batch sizes found)
            for tr in prev_transformer:
                if hasattr(tr, 'batch_size') and isinstance(tr.batch_size, int) and tr.batch_size > batch_size:
                    batch_size = tr.batch_size

        def gen():
            for batch in chunked(iter, batch_size):
                yield from prev_transformer.transform_iter(batch)
        return last_transformer.index(gen()) 

    def transform_iter(self, inp: pt.model.IterDict) -> pt.model.IterDict:
        out = inp
        for transformer in self._transformers:
            out = transformer.transform_iter(out)
        return out
    
    def transform(self, inp : pd.DataFrame) -> pd.DataFrame:
        out = inp
        for m in self._transformers:
            out = m.transform(out)
        return out

    def fit(self, topics_or_res_tr, qrels_tr, topics_or_res_va=None, qrels_va=None):
        """
        This is a default implementation for fitting a pipeline. The assumption is that
        all EstimatorBase be composed with a TransformerBase. It will execute any pre-requisite
        transformers BEFORE executing the fitting the stage.
        """
        for m in self._transformers:
            if isinstance(m, Estimator):
                m.fit(topics_or_res_tr, qrels_tr, topics_or_res_va, qrels_va)
            else:
                topics_or_res_tr = m.transform(topics_or_res_tr)
                # validation is optional for some learners
                if topics_or_res_va is not None:
                    topics_or_res_va = m.transform(topics_or_res_va)

    def __repr__(self):
        return '(' + ' >> '.join([str(t) for t in self._transformers]) + ')'

    def compile(self, verbose: bool = False) -> Transformer:
        """Returns a new transformer that iteratively fuses adjacent transformers to form a more efficient pipeline."""
        # compile constituent transformers (flatten allows compile() to return Compose pipelines)
        inp = deque(_flatten((t.compile() for t in self._transformers), Compose))
        out : deque = deque()
        counter = 1
        while inp:
            if verbose:
                print(counter, list(out), list(inp))
            counter +=1 
            right = inp.popleft()
            if out and isinstance(out[-1], SupportsFuseRight) and (fused := out[-1].fuse_right(right)) is not None:
                if verbose:
                    print(f"  fuse_right {out[-1]} >> {right}  == {fused}")
                out.pop()
                # add the fused pipeline to the start of the input queue so it will be processed next (must be done in reverse due to how extendleft works)
                inp.extendleft(reversed(_flatten([fused], Compose)))
            elif out and isinstance(right, SupportsFuseLeft) and (fused := right.fuse_left(out[-1])) is not None:
                if verbose:
                    print(f"  fuse_left {out[-1]} >> {right}  == {fused}")
                out.pop()
                # add the fused pipeline to the start of the input queue so it will be processed next (must be done in reverse due to how extendleft works)
                inp.extendleft(reversed(_flatten([fused], Compose)))
            else:
                out.append(right)
            if counter == MAX_COMPILE_ITER:
                raise OverflowError()
        if len(out) == 1:
            return out[0]
        return Compose(*out)

    def transform_inputs(self):
        # The first transformer in the pipeline may accept multiple input configurations, but not all of these
        # may work for the rest of the pipeline. So find out which (if any) of the input configurations work, and
        # prioritise those.
        io_configurations = [
            {
                'input_columns': input_columns,
                'output_columns': input_columns,
            }
            for input_columns in pt.inspect.transformer_inputs(self._transformers[0])
        ]
        for transformer in self:
            for configuration in io_configurations:
                if configuration['output_columns'] is None:
                    continue
                configuration['output_columns'] = pt.inspect.transformer_outputs(transformer, configuration['output_columns'], strict=False)
        return [io_cfg['input_columns'] for io_cfg in sorted(io_configurations, key=lambda x: x['output_columns'] is None)]

    def transform_outputs(self, input_columns):
        # Figure out the output columns for the given input columns. This is a more direct and robust way of getting the outputs
        # for a composed pipeline than using inspect's default implementation (running an empty dataframe through the whole pipeline)
        # since it can leverage `transform_outputs` implementation of each transformer.
        output_columns = input_columns
        for transformer in self:
            output_columns = pt.inspect.transformer_outputs(transformer, output_columns)
        return output_columns

    def schematic(self, *, input_columns):
        pipeline = []
        columns = input_columns
        for transformer in self:
            schematic = pt.schematic.transformer_schematic(transformer, input_columns=columns)
            pipeline.append(schematic)
            columns = schematic['output_columns']
        return {
            'type': 'pipeline',
            'input_columns': pipeline[0]['input_columns'] if pipeline else None,
            'output_columns': pipeline[-1]['output_columns'] if pipeline else None,
            'title': None,
            'transformers': pipeline,
        }


MAX_COMPILE_ITER = 10_000



import pyterrier as pt

import itertools
import unittest
import tempfile
import shutil
import os

from .base import TempDirTestCase, BaseTestCase

class TestIterDictIndexer(TempDirTestCase):
        

    def _create_index(self, it, indexer):
        print("Writing index to " + self.test_dir)
        indexref = indexer.index(it)
        self.assertIsNotNone(indexref)
        return indexref

    def _make_check_index(self, n, index_type, fields=('text',), meta=('docno', 'url', 'title')):
        assert fields[0] == 'text'
        from pyterrier.terrier.index import IndexingType
        # Test both versions: _fifo (for UNIX) and _nofifo (for Windows)
        indexers = [
            pt.index._IterDictIndexer_fifo(self.test_dir, type=index_type, meta=meta),
            pt.index._IterDictIndexer_fifo(self.test_dir, type=index_type, threads=4, meta=meta),
            pt.index._IterDictIndexer_nofifo(self.test_dir, type=index_type, meta=meta),
        ]
        if pt.utils.is_windows():
           indexers = [indexers[-1]] 
        for indexer in indexers:
            with self.subTest(indexer=indexer):
                it = (
                    {'docno': '1', 'url': 'url1', 'text': 'He ran out of money, so he had to stop playing', 'title': 'Woes of playing poker', 'raw_source': b'<some>xml</content>'},
                    {'docno': '2', 'url': 'url2', 'text': 'The waves were crashing on the shore; it was a', 'title': 'Lovely sight', 'raw_source': b'<some>xml</content>'},
                    {'docno': '3', 'url': 'url3', 'text': 'The body may perhaps compensates for the loss', 'title': 'Best of Viktor Prowoll', 'raw_source': b'<some>xml</content>'},
                )
                it = itertools.islice(it, n)
                indexref = self._create_index(it, indexer)
                index = pt.IndexFactory.of(indexref)
                self.assertIsNotNone(index)
                self.assertEqual(n, index.getCollectionStatistics().getNumberOfDocuments())
                self.assertEqual(len(set(meta)), len(index.getMetaIndex().getKeys()))
                for m in meta:
                    self.assertTrue(m in index.getMetaIndex().getKeys())
                if index_type is IndexingType.CLASSIC:
                    self.assertTrue(os.path.isfile(self.test_dir + '/data.direct.bf'))
                    self.assertTrue(index.hasIndexStructure("direct"))
                else:
                    self.assertFalse(os.path.isfile(self.test_dir + '/data.direct.bf'))
                inv = index.getInvertedIndex()
                lex = index.getLexicon()
                post = inv.getPostings(lex.getLexiconEntry('plai'))
                post.next()
                self.assertEqual(1, post.frequency)
                index.close()
            # reset index directory for next run
            shutil.rmtree(self.test_dir)
            os.mkdir(self.test_dir)


    def test_checkjavaDocIterator(self):
        it = [
            {'docno': '1', 'url': 'url1', 'text': 'He ran out of money, so he had to stop playing', 'title': 'Woes of playing poker'},
            {'docno': '2', 'url': 'url2', 'text': 'The waves were crashing on the shore; it was a', 'title': 'Lovely sight'},
            {'docno': '3', 'url': 'url3', 'text': 'The body may perhaps compensates for the loss', 'title': 'Best of Viktor Prowoll'},
        ]

        from pyterrier.terrier.index import FlatJSONDocumentIterator

        it1 = itertools.islice(it, 1)
        jIter1 = FlatJSONDocumentIterator(it1)
        self.assertTrue(jIter1.hasNext())
        self.assertIsNotNone(jIter1.next())
        self.assertFalse(jIter1.hasNext())

        it2 = itertools.islice(it, 2)
        jIter1 = FlatJSONDocumentIterator(it2)
        self.assertTrue(jIter1.hasNext())
        self.assertIsNotNone(jIter1.next())
        self.assertTrue(jIter1.hasNext())
        self.assertIsNotNone(jIter1.next())
        self.assertFalse(jIter1.hasNext())

    def test_createindex1(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(1, IndexingType.CLASSIC)

    def test_createindex2(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(2, IndexingType.CLASSIC)

    def test_createindex3(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(3, IndexingType.CLASSIC)

    def test_createindex1_single_pass(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(1, IndexingType.SINGLEPASS)

    def test_createindex2_single_pass(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(2, IndexingType.SINGLEPASS)

    def test_createindex3_single_pass(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(3, IndexingType.SINGLEPASS)

    # Issue #43

    # def test_createindex3_memory(self):
    #     from pyterrier.terrier.index import IndexingType
    #     self._make_check_index(3, IndexingType.MEMORY)

    # def test_createindex1_2fields_memory(self):
    #     from pyterrier.terrier.index import IndexingType
    #     self._make_check_index(1, IndexingType.MEMORY, fields=['text', 'title'])

    def test_long_tokens(self):
        longword = 'abc' * (1+int(60/3))
        indexer = pt.IterDictIndexer(self.test_dir, tokeniser='identity', overwrite=True, properties={'max.term.length' : '80'})
        ref = indexer.index([ {"docno" : "d1", 'text': longword}])
        index = pt.IndexFactory.of(ref)
        self.assertEqual(1, len(index))
        self.assertEqual(1, index.getCollectionStatistics().getNumberOfUniqueTerms())
        self.assertEqual('80', pt.java.cast("org.terrier.structures.PropertiesIndex", index).getIndexProperty("max.term.length", None))
        self.assertTrue(longword in index.getLexicon())

    def test_meta_init(self):
        it = [
            {'docno': '11', 'url': 'url1', 'text': 'He ran out of money, so he had to stop playing', 'title': 'Woes of playing poker'},
            {'docno': '12', 'url': 'url2', 'text': 'The waves were crashing on the shore; it was a', 'title': 'Lovely sight'},
            {'docno': '13', 'url': 'url3', 'text': 'The body may perhaps compensates for the loss', 'title': 'Best of Viktor Prowoll'},
        ]
        props={}
        props["termpipelines"] = ""
        
        indexer = pt.IterDictIndexer(self.test_dir, meta={'docno' : 10, 'url' : 10, 'text' : 100, 'title' : 100}, meta_reverse=['docno', 'url'])
        indexref = indexer.index(it)
        index = pt.IndexFactory.of(indexref)
        self.assertIn("docno", index.getMetaIndex().getKeys())
        self.assertIn("text", index.getMetaIndex().getKeys())
        self.assertIn("docno", index.getMetaIndex().getKeys())
        self.assertIn("url", index.getMetaIndex().getReverseKeys())

        indexer = pt.IterDictIndexer(self.test_dir, meta={'docno' : 10, 'url' : 10, 'text' : 100, 'title' : 100, 'missing_field': 100}, meta_reverse=['docno', 'url'])
        with self.assertRaises(ValueError):
            indexer.index(it) # missing field

        indexer = pt.IterDictIndexer(self.test_dir, meta={'docno' : 1, 'url' : 10, 'text' : 100, 'title' : 100}, meta_reverse=['docno', 'url'])
        with self.assertRaises(ValueError):
            indexer.index(it) # docno not long enough

    def test_check_nostemmer_utf_toks(self):
        # check that UTFTokeniser is doing its job
        it = [
            {'docno': '3', 'url': 'url3', 'text': 'The body may perhaps compensates før the loss'},
        ]
        
        indexer = pt.IterDictIndexer(self.test_dir, stemmer=None, stopwords=None, tokeniser="UTFTokeniser")
        indexref = indexer.index(it)
        index = pt.IndexFactory.of(indexref)
        index = pt.java.cast("org.terrier.structures.IndexOnDisk", index)
        self.assertEqual("", index.getIndexProperty("termpipelines", "bla"))
        self.assertTrue("før" in index.getLexicon())
        # restore setting after test
        pt.terrier.J.ApplicationSetup.setProperty("termpipelines", "Stopwords,PorterStemmer")

    def test_check_customstops(self):
        if not pt.terrier.check_version("5.8"):
            self.skipTest("Requires Terrier 5.8")
        it = [
            {'docno': '1', 'url': 'url1', 'text': 'He ran out of møney, so he had to stop playing', 'title': 'Woes of playing poker'},
            {'docno': '2', 'url': 'url2', 'text': 'The waves were crashing on the shore; it was a', 'title': 'Lovely sight'},
            {'docno': '3', 'url': 'url3', 'text': 'The body may perhaps compensates før the loss', 'title': 'Best of Viktor Prowoll'},
        ]
        indexer = pt.IterDictIndexer(self.test_dir, tokeniser=pt.TerrierTokeniser.utf, stemmer=pt.TerrierStemmer.none, stopwords=['møney', 'crashing', ','], overwrite=True)
        indexref = indexer.index(it)
        index = pt.IndexFactory.of(indexref)
        index = pt.java.cast("org.terrier.structures.IndexOnDisk", index)
        self.assertIn(member="PyTerrierCustomStopwordList$Retrieval", container=index.getIndexProperty("termpipelines", "bla"))
        self.assertIsNotNone(index.getIndexProperty('pyterrier.stopwords', None))

        # lets validate the property
        self.assertTrue("møney" in index.getIndexProperty('pyterrier.stopwords', None))
        self.assertTrue("crashing" in index.getIndexProperty('pyterrier.stopwords', None))

        # lets validate the actual stopword list as parsed from the property
        actual_stopwords = pt.java.autoclass("org.terrier.python.PyTerrierCustomStopwordList$Retrieval")()
        actual_stopwords.setIndex(index)
        self.assertEqual(3, actual_stopwords.stopWords.size())
        self.assertTrue(actual_stopwords.stopWords.contains(","))
        self.assertTrue(actual_stopwords.stopWords.contains("crashing"))
        self.assertTrue(actual_stopwords.stopWords.contains("møney"))

        # lets validate that the index is correct
        self.assertEqual(10, index.getDocumentIndex().getDocumentLength(0)) # møney removed?
        
        # lets validate retrieval and the rest of the index
        self.assertFalse("møney" in index.getLexicon())
        br = pt.terrier.Retriever(index)
        res = br.search("crashing")
        self.assertEqual(0, len(res))
        res = br.search("møney")
        self.assertEqual(0, len(res))
        
    def test_check_nostemmer(self):
        it = [
            {'docno': '1', 'url': 'url1', 'text': 'He ran out of money, so he had to stop playing', 'title': 'Woes of playing poker'},
            {'docno': '2', 'url': 'url2', 'text': 'The waves were crashing on the shore; it was a', 'title': 'Lovely sight'},
            {'docno': '3', 'url': 'url3', 'text': 'The body may perhaps compensates før the loss', 'title': 'Best of Viktor Prowoll'},
        ]
        settings = [( pt.TerrierStemmer.none, pt.TerrierStopwords.none), (None, None)]
        for setting in settings:
            with self.subTest('setting %s' % str(setting)):
                indexer = pt.IterDictIndexer(self.test_dir, stemmer=setting[0], stopwords=setting[1], overwrite=True)
                indexref = indexer.index(it)
                index = pt.IndexFactory.of(indexref)
                index = pt.java.cast("org.terrier.structures.IndexOnDisk", index)
                self.assertEqual("", index.getIndexProperty("termpipelines", "bla"))
                self.assertEqual(11, index.getDocumentIndex().getDocumentLength(0))
                
                # før is tokenised as f r by EnglishTokeniser
                self.assertTrue("f" in index.getLexicon())
                self.assertTrue("r" in index.getLexicon())

                # restore setting after test
                pt.terrier.J.ApplicationSetup.setProperty("termpipelines", "Stopwords,PorterStemmer")
                self.assertEqual("", index.getIndexProperty("termpipelines", "bla"))

    def test_check_weakstemmer(self):
        it = [
            {'docno': '1', 'url': 'url1', 'text': 'He ran out of money, so he had to stop playing', 'title': 'Woes of playing poker'},
            {'docno': '2', 'url': 'url2', 'text': 'The waves were crashing on the shore; it was a', 'title': 'Lovely sight'},
            {'docno': '3', 'url': 'url3', 'text': 'The body may perhaps compensates for the loss', 'title': 'Best of Viktor Prowoll'},
        ]
        settings = ['WeakPorterStemmer', 'weakporter', pt.TerrierStemmer.weakporter]
        for setting in settings:
            with self.subTest('setting %s' % str(setting)):
                indexer = pt.IterDictIndexer(self.test_dir, stemmer=setting, overwrite=True)
                indexref = indexer.index(it)
                index = pt.IndexFactory.of(indexref)
                self.assertEqual(3, index.getDocumentIndex().getDocumentLength(0)) #ran money playing
                index = pt.java.cast("org.terrier.structures.IndexOnDisk", index)
                # restore setting after test
                pt.terrier.J.ApplicationSetup.setProperty("termpipelines", "Stopwords,PorterStemmer")
                self.assertEqual("Stopwords,WeakPorterStemmer", index.getIndexProperty("termpipelines", "bla"))

if __name__ == "__main__":
    unittest.main()



import ast
import pyterrier as pt
from docutils import nodes
from docutils.parsers.rst import Directive, directives


def run_and_return_last(code: str, globals_dict=None, locals_dict=None, source=None, lineno=None):
    if globals_dict is None:
        globals_dict = {}
    if locals_dict is None:
        locals_dict = globals_dict

    # Parse the code into an AST
    parsed = ast.parse(code, mode="exec")

    if source is not None:
        source = source+f":{lineno}"

    # If the last node is an expression, we separate it
    *body, last = parsed.body
    try:
        if isinstance(last, ast.Expr):
            # Make a new module for the body and compile + exec it
            module_body = ast.Module(body=body, type_ignores=[])
            compiled_body = compile(module_body, filename="<input>", mode="exec")
            exec(compiled_body, globals_dict, locals_dict)

            # Compile just the expression part and eval it
            expr = ast.Expression(last.value)
            compiled_expr = compile(expr, filename="<input>", mode="eval")
            return eval(compiled_expr, globals_dict, locals_dict)
        else:
            # No final expression -> just execute the whole thing
            exec(compile(parsed, filename="<input>", mode="exec"), globals_dict, locals_dict)
            return None
    except Exception as e:
        if source is None:
            raise RuntimeError(f"Error executing code: {code} because {e.args}") from e
        raise RuntimeError(f"Error executing code: {code} from {source} because {e.args}") from e


class SchematicDirective(Directive):
    required_arguments = 0
    has_content = True
    option_spec = {'input_columns': str, 'show_code': directives.flag}

    def run(self):
        # File name of the current .rst document
        source = self.reporter.source  # absolute path to the .rst file

        # Line number where the directive was invoked
        lineno = self.state_machine.abs_line_number()  # Directive base class stores this

        code = "\n".join(self.content)
        local_dict = {}
        result = run_and_return_last(code, {'pt': pt}, local_dict, source=source, lineno=lineno)
        if result is None:
            for local_var in ['pipeline', 'transformer']:
                if local_var in local_dict:
                    result = local_dict[local_var]
                    break
        if not isinstance(result, (dict, pt.Transformer)):
            return [self.state_machine.reporter.error(f"Expected dict or Transformer, got {result!r} (type: {type(result)})", line=self.lineno)]
        # parse any input columns supplied in the directive 
        input_columns = self.options.get('input_columns')
        if input_columns is not None:
            input_columns = [x.strip() for x in input_columns.split(',')]

        html = pt.schematic.draw(result, input_columns=input_columns)

        result = [nodes.raw('', html, format='html')]

        if 'show_code' in self.options:
            # prepend a .. code-block:: python
            # To hide parts of the code that are not important, users can include "# FOLD". Only the content after the fold is shown.
            code_to_present = code.split('# FOLD')[-1].strip()
            literal = nodes.literal_block(code_to_present, code_to_present)
            literal['language'] = 'python'
            literal['highlight_args'] = {}
            result = [literal] + result

        return result


def setup(app):
    app.add_directive('schematic', SchematicDirective)
    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }
